import vmi


def view_surface(view, image, iso_value):
    pd = vmi.imIsosurface(image, iso_value)  # 提取CT值为iso_value的等值面，面绘制三维重建
    actor = vmi.PolyActor(view)  # 面绘制显示
    actor.clone(pd)  # 载入三维重建得到的面网格模型
    view.cameraFit()  # 自动调整视图的视野范围
    return actor


def view_volume(view, image, iso_value):
    volume = vmi.ImageVolume(view)  # 体绘制显示
    volume.clone(image)  # 载入读取到的DICOM图像
    volume.opacityScalar({iso_value - 1: 0, iso_value: 1})  # 设置像素和透明度之间的函数关系
    view.cameraFit()  # 自动调整视图的视野范围
    return view, volume


if __name__ == '__main__':
    from PySide2.QtWidgets import QWidget, QHBoxLayout
    from vmi.example.read_dicom_slice_view import read_dicom

    image = read_dicom()  # 读取DICOM路径
    if image is None:
        vmi.appexit()

    iso_value = vmi.askInt(-1000, 200, 3000, None, 'HU')  # 用户输入-1000HU至3000HU范围内的一个CT值
    if iso_value is None:
        vmi.appexit()

    surface_view = vmi.View()  # 面绘制视图
    volume_view = vmi.View()  # 体绘制视图

    view_surface(surface_view, image, iso_value)
    view_volume(volume_view, image, iso_value)

    # 创建窗口包含两个视图
    widget = QWidget()
    layout = QHBoxLayout(widget)
    layout.addWidget(surface_view)
    layout.addWidget(volume_view)

    vmi.appexec(widget)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



from typing import List, Dict, Tuple, AnyStr
import threading
import numpy as np
import time

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from PySide2.QtWinExtras import *

import vtk

import vmi

tr = QObject()
tr = tr.tr


class SelectEdit(QObject, vmi.Menu, vmi.Mouse):
    def __init__(self, view: vmi.View, name=None):
        QObject.__init__(self)

        self.name = name if name else tr('选区编辑 (Select edit)')
        vmi.Menu.__init__(self, name=self.name)
        self.actions = {'Activate': QAction(tr('激活 (Activate)')),
                        'Cancel': QAction(tr('取消 (Cancel)')),
                        'Apply': QAction(tr('应用 (Apply)')),
                        'CurveType': QAction('')}

        self.actions['Activate'].triggered.connect(self.activate)
        self.actions['Cancel'].triggered.connect(self.cancel)
        self.actions['Apply'].triggered.connect(self.apply)
        self.actions['CurveType'].triggered.connect(self.setCurveType)

        def aboutToShow():
            self.actions['Activate'].setEnabled(not self.isActivated())
            self.actions['Cancel'].setEnabled(self.isActivated())
            self.actions['Apply'].setEnabled(self.canApply())
            self.actions['CurveType'].setText('{} = {}'.format(tr('线型 (Curve type)'),
                                                               self.curveTypeText[self.curveType]))

            self.menu.clear()
            self.menu.addAction(self.actions['Activate'])
            self.menu.addAction(self.actions['Cancel'])
            self.menu.addSeparator()
            self.menu.addAction(self.actions['CurveType'])
            self.menu.addSeparator()
            self.menu.addAction(self.actions['Apply'])

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self, menu=self)
        self.mouse['LeftButton']['PressMove'] = self.mouseblock
        self.mouse['LeftButton']['PressRelease'] = self.pick

        self.view = view

        self.nodes: List[Tuple] = []
        self.nodeActors: List[vmi.PolyActor] = []
        self.nodeShapes: List[vmi.ShapeData] = []

        self.curveShape = vmi.ShapeData()
        self.curveActor = vmi.PolyActor(self.view)
        self.curveActor.color(rgb=(1, 0.2, 0.2))
        self.curveActor.size(line=4)
        self.curveActor.alwaysOnTop(True)
        self.curveActor.pickable(True)
        self.curveActor.bind(self.curveShape)
        self.curveActor.mouse['LeftButton']['PressMove'] = self.move

        self.curveType = 0
        self.curveTypeText = dict(enumerate([tr('折线 (Polyline)'), tr('样条 (Spline)')]))

        self.assignDefault = -3071
        self._Bind: vmi.ImageSlice = None

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]
        return s

    def move(self, **kwargs):
        if kwargs['picked'] in self.nodeActors:
            over = self.view.mouseOverFPlane()
            j = self.nodeActors.index(kwargs['picked'])
            node = [self.nodes[j][i] + over[i] for i in range(3)]
            if node not in self.nodes:
                self.nodes[j] = node
                self.rebuild()
        elif kwargs['picked'] is self.curveActor:
            over = self.view.mouseOverFPlane()
            for j in range(len(self.nodes)):
                for i in range(3):
                    self.nodes[j][i] += over[i]
            self.rebuild()

    def pick(self, **kwargs):
        if kwargs['double']:
            if kwargs['picked'] in self.nodeActors:
                j = self.nodeActors.index(kwargs['picked'])
                self.removeNode(self.nodeActors[j])
            else:
                pt = self.view.pickFPlane()
                if self._Bind is not None:
                    cnt = vmi.imCenter(self._Bind.data())
                    look = self.view.camera('look')
                    pt = vmi.ptOnPlane(pt, cnt, look)
                self.addNode(pt)

    def nodeMenu(self, **kwargs):
        if kwargs['picked'] in self.nodeActors:
            i = self.nodeActors.index(kwargs['picked'])
            pos = self.view.pickToDisplay(self.nodes[i])
            pos = (pos.x(), pos.y())

            menu = QMenu()
            nodeReset = menu.addAction(tr('坐标 (Coordinates)') + ' = ' + repr(pos))

            action = menu.exec_(QCursor.pos())

            if action is nodeReset:
                size = (self.view.width(), self.view.height())
                pos = vmi.askInts([0, 0], pos, size, '{} {}'.format(
                    tr('坐标 (Coordinates)'), repr(size)))
                if pos is not None:
                    node = self.view.pickFPlane(QPoint(pos[0], pos[1]))
                    if node not in self.nodes:
                        self.nodes[i] = node
                        self.rebuild()

    def bind(self, bind):
        if self._Bind is not bind:
            self._Bind = bind

    def setCurveType(self, *args, lineType=None):
        if lineType is None:
            self.curveType += 1
            self.curveType = self.curveType % len(self.curveTypeText)
            self.rebuild()
        elif lineType.lower() == 'polyline':
            self.curveType = 0
            self.rebuild()
        elif lineType.lower() == 'spline':
            self.curveType = 1
            self.rebuild()

    def activate(self):
        self.view.addMouse(self)

    def isActivated(self):
        return self.view.checkMouse(self)

    def canApply(self):
        return bool(self._Bind) and len(self.nodes) > 2

    def apply(self):
        if self._Bind is not None:
            target = vmi.askInt(vtk.VTK_SHORT_MIN, self.assignDefault, vtk.VTK_SHORT_MAX, tr('赋值 (Assign value)'))

            face = vmi.ccFace(self.curveShape.shape())
            look = self.view.camera('look')
            lthn = vmi.imSize_Vt(self._Bind.data(), self.view.camera('look'))
            region = vmi.ccPrism(face, [-look[i] for i in range(3)], lthn)
            region = vmi.ccBoolean_Union(vmi.ccPrism(face, look, lthn), region)
            region = vmi.ccPd_Sh(region)

            stencil = vmi.imStencil(self._Bind.data(), region)
            ext = vmi.imExtent(self._Bind.data(), region.GetBounds())
            it = vmi.imIterator(self._Bind.data(), ext, stencil)

            p = [0]

            def func():
                count, sum = 0, (ext[5] - ext[4] + 1) * (ext[3] - ext[2] + 1) * (ext[1] - ext[0] + 1)
                while not it.IsAtEnd():
                    if it.IsInStencil():
                        vmi.imValue(self._Bind.data(), it.GetIndex(), target)
                    it.Next()
                    count += 1
                    p[0] = 100 * count / sum

            t = threading.Thread(target=func)
            t.start()

            vmi.appwait(t, p)
            self.view.updateInTime()

    def cancel(self):
        self.view.removeMouse(self)
        self.clearNode()
        self.rebuild()

    def addNode(self, pt):
        if pt not in self.nodes:
            self.nodes.append(pt)
            self.nodeShapes.append(vmi.ShapeData())
            self.nodeActors.append(vmi.PolyActor(self.view))
            self.nodeActors[-1].color(rgb=(0.2, 0.6, 1))
            self.nodeActors[-1].size(point=8)
            self.nodeActors[-1].pickable(True)
            self.nodeActors[-1].alwaysOnTop(True)
            self.nodeActors[-1].bind(self.nodeShapes[-1])
            self.nodeActors[-1].mouse['LeftButton']['PressMove'] = self.move
            self.nodeActors[-1].mouse['LeftButton']['PressRelease'] = self.pick
            self.nodeActors[-1].mouse['RightButton']['PressRelease'] = self.nodeMenu
            self.rebuild()

    def removeNode(self, actorNode: vmi.PolyActor):
        actorNode.delete()
        i = self.nodeActors.index(actorNode)
        del self.nodes[i], self.nodeShapes[i], self.nodeActors[i]
        self.rebuild()

    def clearNode(self):
        for actorNode in self.nodeActors:
            actorNode.delete()
        self.nodes.clear()
        self.nodeShapes.clear()
        self.nodeActors.clear()
        self.rebuild()

    def rebuild(self):
        for i in range(len(self.nodes)):
            self.nodeShapes[i].clone(vmi.ccVertex(vmi.ccPnt(self.nodes[i])))
        if len(self.nodes) > 2:
            if self.curveType == 0:
                self.curveShape.clone(vmi.ccWire(vmi.ccSegments(pts=self.nodes, closed=True)))
            elif self.curveType == 1:
                self.curveShape.clone(vmi.ccWire(vmi.ccBSpline_Kochanek(pts=self.nodes, closed=True)))
        elif len(self.nodes) > 1:
            if self.curveType == 0:
                self.curveShape.clone(vmi.ccWire(vmi.ccSegments(pts=self.nodes, closed=False)))
            elif self.curveType == 1:
                self.curveShape.clone(vmi.ccWire(vmi.ccBSpline_Kochanek(pts=self.nodes, closed=False)))
        else:
            self.curveShape.clone(None)
        self.curveActor.view.updateInTime()


class RegionResample(QObject, vmi.Menu, vmi.Mouse):
    def __init__(self, view: vmi.View, name=None):
        QObject.__init__(self)

        self.name = name if name else tr('区域重采样 (Region resample)')
        vmi.Menu.__init__(self, name=self.name)
        self.actions = {'Activate': QAction(tr('激活 (Activate)')),
                        'Cancel': QAction(tr('取消 (Cancel)')),
                        'Apply': QAction(tr('应用 (Apply)')),
                        'SetBnd': QAction(tr('输入边界 (Input bounds)')),
                        'ResetBnd': QAction(tr('重置边界 (Reset bounds)'))}

        self.actions['Activate'].triggered.connect(self.activate)
        self.actions['Cancel'].triggered.connect(self.cancel)
        self.actions['Apply'].triggered.connect(self.apply)
        self.actions['SetBnd'].triggered.connect(self.setBnd)
        self.actions['ResetBnd'].triggered.connect(self.resetBnd)

        def aboutToShow():
            self.actions['Activate'].setEnabled(not self.isActivated())
            self.actions['Cancel'].setEnabled(self.isActivated())
            self.actions['Apply'].setEnabled(self.canApply())
            self.actions['ResetBnd'].setEnabled(bool(self._Bind))

            self.menu.clear()
            self.menu.addAction(self.actions['Activate'])
            self.menu.addAction(self.actions['Cancel'])
            self.menu.addSeparator()
            self.menu.addAction(self.actions['SetBnd'])
            self.menu.addAction(self.actions['ResetBnd'])
            self.menu.addSeparator()
            self.menu.addMenu(self.view.menu)
            self.menu.addSeparator()
            self.menu.addAction(self.actions['Apply'])

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self, menu=self)

        self.view = view

        self.regionShapes = []
        self.regionActors = []

        for k in range(3):
            jlist = [], []
            for j in range(2):
                ilist = [], []
                for i in range(2):
                    shape = vmi.ShapeData()
                    actor = vmi.PolyActor(self.view)
                    actor.color(rgb=[1, 0.2, 0.2])
                    actor.size(line=4)
                    actor.alwaysOnTop(True)
                    actor.pickable(True)
                    actor.bind(shape)
                    actor.mouse['LeftButton']['PressMove'] = self.move

                    ilist[0].append(shape)
                    ilist[1].append(actor)
                jlist[0].append(ilist[0])
                jlist[1].append(ilist[1])
            self.regionShapes.append(jlist[0])
            self.regionActors.append(jlist[1])

        self._Bind: vmi.ImageSlice = None
        self._Bnd = [0, 1, 0, 1.5, 0, 2]

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]
        return s

    def activate(self):
        self.view.addMouse(self)
        self.rebuild()

    def isActivated(self):
        return self.view.checkMouse(self)

    def canApply(self):
        return bool(self._Bind)

    def apply(self):
        if self._Bind is not None:
            spc = vmi.askFloats([0.1, 0.1, 0.1], vmi.imSpacing_Vt(self._Bind.data()), [5, 5, 5], 3,
                                tr('采样间距 (Sample spacing) (x, y, z) (mm)'))
            if spc is not None:
                ori, dim = [self._Bnd[0], self._Bnd[2], self._Bnd[4]], [1, 1, 1]
                for i in range(3):
                    dim[i] = int((self._Bnd[2 * i + 1] - self._Bnd[2 * i]) / spc[i])
                ext = [0, dim[0] - 1, 0, dim[1], 0, dim[2]]
                self._Bind.clone(vmi.imResample(self._Bind.data(), ori, ext, spc))
                self.view.updateInTime()

    def cancel(self):
        self.view.removeMouse(self)
        self.rebuild()

    def move(self, **kwargs):
        picked, k, j, i = False, 0, 0, 0
        for k in range(3):
            for j in range(2):
                for i in range(2):
                    if kwargs['picked'] is self.regionActors[k][j][i]:
                        picked = True
                        break
                if picked is True:
                    break
            if picked is True:
                break
        if picked:
            bnd = [[self._Bnd[0], self._Bnd[2], self._Bnd[4]],
                   [self._Bnd[1], self._Bnd[3], self._Bnd[5]]]
            for delta in range(1, 3):
                a, b = (k + delta) % 3, [k, j, i][delta]
                vt = [0, 0, 0]
                vt[a] = 1
                over = self.view.mouseOverFPlane(vt)
                over = vmi.signNorm3(over, vt)
                bnd[b][a] += over
            self._Bnd = [bnd[0][0], bnd[1][0], bnd[0][1], bnd[1][1], bnd[0][2], bnd[1][2]]
            self.rebuild()
        else:
            return 'pass'

    def bind(self, bind):
        if self._Bind is not bind:
            self._Bind = bind
            self.resetBnd()

    def setBnd(self, *args, bnd=None):
        if bnd is None:
            imbnd = vmi.imBounds(self._Bind.data()) if self._Bind else [vtk.VTK_SHORT_MIN, vtk.VTK_SHORT_MAX] * 3
            bnd = vmi.askFloats([imbnd[0], imbnd[0], imbnd[2], imbnd[2], imbnd[4], imbnd[4]],
                                [self._Bnd[0], self._Bnd[1], self._Bnd[2], self._Bnd[3], self._Bnd[4], self._Bnd[5]],
                                [imbnd[1], imbnd[1], imbnd[3], imbnd[3], imbnd[5], imbnd[5]], 3,
                                tr('边界 (Bounds) = (xmin, xmax, ymin, ymax, zmin, zmax) (mm)'))
        if bnd is not None:
            if bnd[0] > bnd[1] or bnd[2] > bnd[3] or bnd[4] > bnd[5]:
                vmi.askInfo(tr('输入无效 (Invalid input)'))
                return
            self._Bnd = bnd
            self.rebuild()

    def resetBnd(self, *args):
        if self._Bind is not None:
            self._Bnd = vmi.imBounds(self._Bind.data())
            self.rebuild()

    def rebuild(self):
        if self.isActivated():
            for k in range(3):
                for j in range(2):
                    for i in range(2):
                        bnd = [[self._Bnd[0], self._Bnd[2], self._Bnd[4]],
                               [self._Bnd[1], self._Bnd[3], self._Bnd[5]]]
                        pt = [[0, 0, 0], [0, 0, 0]]
                        a, b = (k + 1) % 3, (k + 2) % 3

                        pt[0][k], pt[1][k] = bnd[0][k], bnd[1][k]
                        pt[0][a] = pt[1][a] = bnd[j][a]
                        pt[0][b] = pt[1][b] = bnd[i][b]
                        self.regionShapes[k][j][i].clone(vmi.ccEdge(vmi.ccSegment(pt[0], pt[1])))
        else:
            for k in range(3):
                for j in range(2):
                    for i in range(2):
                        self.regionShapes[k][j][i].clone(None)
        self.view.updateInTime()




import sys
from typing import List, Optional, Union

import vmi
import vtk
import numpy as np
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtOpenGL import QGLWidget
from PySide2.QtWidgets import *

vtk.vtkObject().SetGlobalWarningDisplay(0)


class Mouse:
    def __init__(self, menu=None):
        i = 0

        def e():
            nonlocal i
            name = '_MouseEvent{}'.format(i)
            if not hasattr(self, name):
                setattr(self, '_MouseEvent{}'.format(i), self.mousepass)
                i += 1
            return getattr(self, name)

        self.mouse = {
            'NoButton': {'Enter': e(), 'Leave': e(), 'Wheel': e(), 'Move': e()},
            'LeftButton': {'Press': e(), 'PressMove': e(), 'PressMoveRelease': e(), 'PressRelease': e()},
            'MidButton': {'Press': e(), 'PressMove': e(), 'PressMoveRelease': e(), 'PressRelease': e()},
            'RightButton': {'Press': e(), 'PressMove': e(), 'PressMoveRelease': e(), 'PressRelease': e()}}

        if menu is not None:
            self.mouse['RightButton']['PressRelease'] = menu.menuexec

    @staticmethod
    def mousepass(**kwargs):
        return 'pass'

    @staticmethod
    def mouseblock(**kwargs):
        return


class Menu:
    def __init__(self, name=None):
        self.menu = QMenu()

        if name is not None:
            self.menu.setTitle(name)

    def menuexec(self, **kwargs):
        self.menu.exec_(QCursor.pos())


class View(QGLWidget, Menu, Mouse):
    """vtk场景"""

    def __init__(self, name=None, parent=None):
        QGLWidget.__init__(self, parent)

        self.name = name if name else self.tr('视图 (View)')
        Menu.__init__(self, name=self.name)

        self.actions = {'CameraFit': QAction(self.tr('最适 (Fit)')),
                        'CameraReverse': QAction(self.tr('反向 (Reverse)')),
                        'CameraAxial': QAction(self.tr('横断位 (Axial)')),
                        'CameraSagittal': QAction(self.tr('矢状位 (Sagittal)')),
                        'CameraCoronal': QAction(self.tr('冠状位 (Coronal)')),
                        'Snapshot': QAction(self.tr('快照 (Snapshot)')), }

        self.actions['CameraFit'].triggered.connect(self.cameraFit)
        self.actions['CameraReverse'].triggered.connect(self.cameraReverse)
        self.actions['CameraAxial'].triggered.connect(self.cameraAxial)
        self.actions['CameraSagittal'].triggered.connect(self.cameraSagittal)
        self.actions['CameraCoronal'].triggered.connect(self.cameraCoronal)
        self.actions['Snapshot'].triggered.connect(self.snapshot)

        self.menu.addAction(self.actions['CameraFit'])
        self.menu.addAction(self.actions['CameraReverse'])
        self.menu.addAction(self.actions['CameraAxial'])
        self.menu.addAction(self.actions['CameraSagittal'])
        self.menu.addAction(self.actions['CameraCoronal'])
        self.menu.addAction(self.actions['Snapshot'])

        Mouse.__init__(self, menu=self)
        self.mouse['LeftButton']['PressMove'] = self.mouseRotateFocal
        self.mouse['MidButton']['PressMove'] = self.mousePan
        self.mouse['RightButton']['PressMove'] = self.mouseZoom
        self._MStack: List[Mouse] = []

        rect = QGuiApplication.screens()[0].geometry()
        rect = QRect(rect.center() - QPoint(800, 450), QSize(1600, 900))
        self.setGeometry(rect)
        self.setAttribute(Qt.WA_OpaquePaintEvent)
        self.setAttribute(Qt.WA_PaintOnScreen)
        self.setFocusPolicy(Qt.WheelFocus)
        self.setMouseTracking(True)

        # 渲染器
        self._Renderer = vtk.vtkRenderer()
        self._Renderer.SetUseHiddenLineRemoval(1)  # 隐线消除
        self._Renderer.SetUseFXAA(False)  # 快速近似抗锯齿
        self._Renderer.SetBackground(1, 1, 1)

        # 渲染窗口
        self._RenderWindow = vtk.vtkRenderWindow()
        self._RenderWindow.SetWindowInfo(str(int(self.winId())))
        self._RenderWindow.AddRenderer(self._Renderer)

        # 交互器
        self._Interactor = vtk.vtkGenericRenderWindowInteractor()
        self._Interactor.SetRenderWindow(self._RenderWindow)
        self._Interactor.Initialize()

        # 方位标
        self._OrientationMarker = init_orientation_marker(self._Interactor)

        # 照相机
        self._Camera = self.camera()
        self.camera().SetParallelProjection(True)  # 平行投影

        # 拾取器
        self._Picker = vtk.vtkPicker()
        self._PickerProp = vtk.vtkPropPicker()
        self._PickerPoint = vtk.vtkPointPicker()
        self._PickerCell = vtk.vtkCellPicker()
        self._Picker.SetTolerance(0.005)
        self._PickerPoint.SetTolerance(0.005)
        self._PickerCell.SetTolerance(0.005)

        # 刷新计时器
        self._Timer = QTimer(self)
        self._Timer.timeout.connect(self.timeout)
        self._Timer.start(10)

        self._Time = QTime()
        self._Time.start()
        self._FPS = self._UpdateSum = 0

        # 创建一个隐藏控件，当销毁时调用清理vtk元素
        self._Hidden = QWidget(self)
        self._Hidden.hide()
        self._Hidden.destroyed.connect(self.delete)

        # 更新请求
        self._UpdateAtOnce: bool = False

        # 鼠标位置
        self._DblClick = False
        self._MousePosLast: QPoint = QPoint(0, 0)
        self._MousePos: QPoint = QPoint(0, 0)

        # 鼠标正在交互的对象
        self.activated = None
        self._PProp: vtk.vtkProp = None
        self._PView: View = None

        # 鼠标按键状态
        self._ButtonPress = self._ButtonPressMove = 'NoButton'

        self.cameraAxial()

    def __setstate__(self, s):
        self.__init__(s['name'])
        self.__dict__.update(s)

        for kw in s['_VTKGets']:
            vmi.vtkset(getattr(self, kw), s['_VTKGets'][kw])

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'menuCamera', 'actions', '_OrientationMarker',
                   '_Timer', '_Hidden', 'aboutToClose', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_VTKGets'] = {name: vmi.vtkget(getattr(self, name)) for name in
                         ['_RenderWindow', '_Renderer', '_Interactor', '_Camera',
                          '_Picker', '_PickerProp', '_PickerPoint', '_PickerCell']}

        for kw in s['_VTKGets']:
            del s[kw]
        return s

    def closeEvent(self, ev):
        self.delete()

    def delete(self):
        self._Timer.stop()
        self._RenderWindow.Finalize()

    def mousexec(self, stack, button, event, picked=None, **kwargs):
        if picked is None:
            if self._PView:
                picked = self._PView
            if self._PProp:
                picked = self._PProp

        for m in [m for m in stack if isinstance(m, Mouse)]:
            if m.mouse[button][event](picked=picked, **kwargs) != 'pass':
                break

    def addMouse(self, mouse):
        if mouse not in self._MStack:
            self._MStack.insert(0, mouse)

    def removeMouse(self, mouse):
        if mouse in self._MStack:
            self._MStack.remove(mouse)

    def checkMouse(self, mouse):
        return mouse in self._MStack and mouse is self._MStack[0]

    def entered(self):
        """返回鼠标悬停状态，在物体上->prop/在场景内->vmi/在场景外->None"""
        return self._PProp if self._PProp else (self._PView if self._PView else self)

    def enterEvent(self, ev):
        self._PView = self
        self._PProp, *_ = self.pickProp(ev.pos())
        self.mousexec([self._PView, self._PProp] + self._MStack, 'NoButton', 'Enter')

    def leaveEvent(self, ev):
        self.mousexec([self._PProp, self._PView] + self._MStack, 'NoButton', 'Leave')
        self._PView = self._PProp = None
        self._ButtonPress = self._ButtonPressMove = 'NoButton'

    def mouseMoveEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if self._ButtonPress == 'NoButton':
            prop, *_ = self.pickProp(ev.pos())
            last_prop, self._PProp = self._PProp, prop

            if not last_prop and not prop:
                self.mousexec(self._MStack + [self._PView], 'NoButton', 'Move')
            elif not last_prop and prop:
                self.mousexec([prop] + self._MStack, 'NoButton', 'Enter', prop)
            elif last_prop and not prop:
                self.mousexec([last_prop] + self._MStack, 'NoButton', 'Leave', last_prop)
            elif last_prop and prop and last_prop is prop:
                self.mousexec([self._PProp] + self._MStack + [self._PView], 'NoButton', 'Move')
            elif last_prop and prop and last_prop is not prop:
                self.mousexec([last_prop] + self._MStack, 'NoButton', 'Leave', last_prop)
                self.mousexec([prop] + self._MStack, 'NoButton', 'Enter', prop)
        elif self._MousePosLast != self._MousePos:
            self._ButtonPressMove = self._ButtonPress
            self.mousexec([self._PProp] + self._MStack + [self._PView], self._ButtonPressMove, 'PressMove')

    def mousePressEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if self._ButtonPress == 'NoButton':
            self._ButtonPress = ev.button().name.decode()
            self._DblClick = ev.type() == QEvent.Type.MouseButtonDblClick
            self.mousexec([self._PProp] + self._MStack + [self._PView], self._ButtonPress, 'Press',
                          double=self._DblClick)

    def mouseReleaseEvent(self, ev):
        self._MousePosLast, self._MousePos = self._MousePos, ev.pos()

        if self._ButtonPressMove == ev.button().name.decode():
            last_button = self._ButtonPress
            self._ButtonPress = self._ButtonPressMove = 'NoButton'
            self.mousexec([self._PProp] + self._MStack + [self._PView], last_button, 'PressMoveRelease')
        elif self._ButtonPress == ev.button().name.decode():
            last_button, last_double = self._ButtonPress, self._DblClick
            self._ButtonPress = self._ButtonPressMove = 'NoButton'
            self._DblClick = False
            self.mousexec([self._PProp] + self._MStack + [self._PView], last_button, 'PressRelease',
                          double=last_double)

    def mouseRotateFocal(self, **kwargs):
        dx, dy = self.mouseOverDisplay()
        rx, ry = -200 * dx / self.height(), 200 * dy / self.width()
        self.camera().Azimuth(rx)
        self.camera().Elevation(ry)
        self.camera().OrthogonalizeViewUp()
        self.updateInTime()

    def mousePan(self, **kwargs):
        over = self.mouseOverFPlane()
        fp, cp = self.camera('focal', 'position')
        fp = [fp[i] - over[i] for i in range(3)]
        cp = [cp[i] - over[i] for i in range(3)]
        self.camera().SetFocalPoint(fp)
        self.camera().SetPosition(cp)

        if self._Interactor.GetLightFollowCamera():
            self._Renderer.UpdateLightsGeometryToFollowCamera()

        self.updateInTime()

    def mouseZoom(self, **kwargs):
        c = self._Renderer.GetCenter()
        _, dy = self.mouseOverDisplay()
        factor = 1.1 ** (10 * -dy / c[1])

        if self.camera().GetParallelProjection():
            self.camera().SetParallelScale(self.camera().GetParallelScale() / factor)
        else:
            self.camera().Dolly(factor)

        if self._Interactor.GetLightFollowCamera():
            self._Renderer.UpdateLightsGeometryToFollowCamera()

        self.updateInTime()

    def mouseOnDisplay(self):
        return self._MousePos.x(), self._MousePos.y()

    def mouseOnFPlane(self):
        return self.pickFPlane(self._MousePos)[1]

    def mouseOnProp(self):
        return self.pick(pos=self._MousePos, picker=self._PickerProp)[0]

    def mouseOnPropCell(self):
        return self.pick(pos=self._MousePos, picker=self._PickerCell)[1]

    def mouseOverDisplay(self):
        """返回鼠标位移向量
        -> (dx, dy)"""
        return self._MousePos.x() - self._MousePosLast.x(), self._MousePos.y() - self._MousePosLast.y()

    def mouseOverFPlane(self, vt=None):
        """返回鼠标在焦平面上的位移向量
        -> (x, y, z)"""
        pt = self.pickFPlane(self._MousePos)
        ptLast = self.pickFPlane(self._MousePosLast)
        over = [pt[i] - ptLast[i] for i in range(3)]
        if vt is not None:
            over = vmi.vtOnVector(over, vt)
        return over

    def mousePos(self):
        """返回当前鼠标位置"""
        return self._MousePos

    def mousePosLast(self):
        """返回上一鼠标位置"""
        return self._MousePosLast

    def paintEngine(self):
        return None

    def paintEvent(self, ev):
        self._Interactor.Render()

    def pick(self, picker, pos: QPoint = None):
        """拾取场景对象"""
        pos = self._MousePos if pos is None else pos

        picker.Pick(pos.x(), self.height() - pos.y(), 0, self._Renderer)
        prop = picker.GetViewProp()
        prop = prop._Prop if prop else None
        return prop, np.array(picker.GetPickPosition()), picker

    def pickBBox(self, pos: QPoint):
        """拾取场景对象的包围盒
        pickBBox(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._Picker)

    def pickCell(self, pos: QPoint):
        """拾取场景对象的单元
        pickCell(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerCell)

    def pickFPlane(self, pos: Union[QPoint, List[int], None] = None):
        """拾取场景的焦平面
        pickFPlane(QPoint(x, y)) -> (x, y, z)"""
        pos = self._MousePos if pos is None else pos
        if not isinstance(pos, QPoint):
            pos = QPoint(pos[0], pos[1])

        t = [0] * 4
        vtk.vtkInteractorObserver.ComputeDisplayToWorld(
            self._Renderer, pos.x(), self.height() - pos.y(), 0, t)
        pt = [_ / t[3] for _ in t[:3]]

        fp, look = self.camera('focal', 'look')
        vtk.vtkPlane.ProjectPoint(pt, fp, look, pt)
        return np.array(pt)

    def pickPoint(self, pos: QPoint):
        """拾取场景对象的点
        pickPoint(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerPoint)

    def pickProp(self, pos: QPoint):
        """拾取场景对象
        pickProp(QPoint(x, y)) -> (x, y, z)"""
        return self.pick(pos=pos, picker=self._PickerProp)

    def pickToDisplay(self, pt):
        """拾取的场景坐标转换为屏幕坐标
        pickToDisplay((x, y, z) -> QPoint(x, y)"""
        t = [0] * 3
        vtk.vtkInteractorObserver.ComputeWorldToDisplay(
            self._Renderer, pt[0], pt[1], pt[2], t)
        return QPoint(t[0], self.height() - t[1])

    def pickToDisplayFlipY(self, pt):
        """拾取的场景坐标转换为屏幕坐标
        pickToDisplay((x, y, z) -> QPoint(x, y)"""
        t = [0] * 3
        vtk.vtkInteractorObserver.ComputeWorldToDisplay(
            self._Renderer, pt[0], pt[1], pt[2], t)
        return QPoint(t[0], t[1])

    def resizeEvent(self, ev):
        w = self.width()
        h = self.height()
        vtk.vtkRenderWindow.SetSize(self._RenderWindow, w, h)
        self._Interactor.SetSize(w, h)
        self._Interactor.ConfigureEvent()
        self.update()

    def camera(self, *args):
        if len(args) == 0:
            return self._Renderer.GetActiveCamera()
        elif len(args) > 0:
            fp = np.array(self.camera().GetFocalPoint())
            cp = np.array(self.camera().GetPosition())
            up = np.array(self.camera().GetViewUp())

            look = fp - cp
            right = np.cross(look, up)
            look /= np.linalg.norm(look)
            right /= np.linalg.norm(right)
            up /= np.linalg.norm(up)

            t, value = [], {'focal': fp, 'position': cp, 'up': up, 'look': look, 'right': right}
            for key in args:
                if key in value.keys():
                    t.append(value[key])
            if len(t) == 1:
                t = t[0]
            return t

    def cameraCS(self):
        focal, right, up, look = self.camera('focal', 'right', 'up', 'look')
        cs = vmi.CS4x4()
        cs.origin(self.pickFPlane(QPoint(0, 0)))
        cs.axis(0, right)
        cs.axis(1, -up)
        cs.axis(2, np.array(look))
        return cs

    def cameraSet(self, face=None, b6=None, focal=None):
        if face is not None:
            face = face.lower()
            vu = {'sagittal': [0, 0, 1],
                  'coronal': [0, 0, 1],
                  'axial': [0, -1, 0]}[face]
            vr = {'sagittal': [0, -1, 0],
                  'coronal': [1, 0, 0],
                  'axial': [1, 0, 0]}[face]
            vl = [0] * 3
            vtk.vtkMath.Cross(vu, vr, vl)

            fp = self.camera('focal')
            d = self.camera().GetDistance()
            cp = [fp[i] - d * vl[i] for i in range(3)]

            self.camera().SetPosition(cp)
            self.camera().SetFocalPoint(fp)
            self.camera().SetViewUp(vu)
            self._Renderer.ResetCameraClippingRange()
        if b6 is not None:
            if isinstance(b6, str) and b6.lower() == 'all':
                b6 = self._Renderer.ComputeVisiblePropBounds()
            b6 = list(b6)

            if focal is not None:
                c = [0.5 * (b6[0] + b6[1]), 0.5 * (b6[2] + b6[3]), 0.5 * (b6[4] + b6[5])]
                c = [focal[i] - c[i] for i in range(3)]
                for i in range(3):
                    if c[i] < 0:
                        b6[2 * i] += 2 * c[i]
                    else:
                        b6[2 * i + 1] += 2 * c[i]

            dxyz = np.array([b6[1] - b6[0], b6[3] - b6[2], b6[5] - b6[4]])
            self._Renderer.ResetCamera(b6)

            vr, vu = self.camera('right', 'up')
            dr = np.abs(vr / np.linalg.norm(vr))
            dr = np.dot(dr, dxyz)
            dr = dr if dr != 0 else 1
            du = np.abs(vu / np.linalg.norm(vu))
            du = np.dot(du, dxyz)
            du = du if du != 0 else 1

            scale = 0.5 * du
            if du / dr < self.height() / self.width():
                scale = 0.5 * dr * self.height() / self.width()
            self.camera().SetParallelScale(scale)
        if focal is not None:
            vl = self.camera('look')
            b6 = self._Renderer.ComputeVisiblePropBounds()
            d = 0.5 * ((b6[1] - b6[0]) ** 2 + (b6[3] - b6[2]) ** 2 + (b6[5] - b6[4]) ** 2) ** 0.5
            cp = [focal[i] - d * vl[i] for i in range(3)]

            self.camera().SetPosition(cp)
            self.camera().SetFocalPoint(focal)
            self._Renderer.ResetCameraClippingRange()
        self.updateInTime()

    def cameraFit(self):
        self.cameraSet(b6='all')

    def cameraReverse(self):
        fp, cp = self.camera('focal', 'position')
        for i in range(3):
            cp[i] = 2 * fp[i] - cp[i]
        self.camera().SetPosition(cp)
        self._Renderer.ResetCameraClippingRange()
        self.updateInTime()

    def cameraAxial(self):
        self.cameraSet(face='axial')

    def cameraSagittal(self):
        self.cameraSet(face='sagittal')

    def cameraCoronal(self):
        self.cameraSet(face='coronal')

    def orientationMarkerVisible(self, arg: bool):
        """显示/隐藏方位标"""
        self._OrientationMarker.GetOrientationMarker().SetVisibility(arg)
        self.updateInTime()

    def sizeHint(self):
        return QSize(1600, 900)

    def snapshot(self, file=None, *args):
        if file is None:
            file = vmi.askSaveFile(self.tr('快照 (Snapshot)'), '*.png')
        if file is not None:
            self.windowHandle().screen().grabWindow(self.winId()).save(file)

    def timeout(self):
        if self._UpdateAtOnce:
            self.updateAtOnce()
        self._UpdateSum += 1

        if self._Time.elapsed() > 0.5e3:
            self._FPS = 1000 * self._UpdateSum / self._Time.elapsed()
            self.setWindowTitle(QApplication.applicationName() + ' FPS: {:.0f}'.format(self._FPS))
            self._Time.restart()
            self._UpdateSum = 0

    def updateInTime(self):
        """及时刷新视图"""
        self._UpdateAtOnce = True

    def updateAtOnce(self):
        if self._Timer.isActive():
            self._UpdateAtOnce = False
            self._Renderer.ResetCameraClippingRange()
            self._RenderWindow.Render()

    def wheelEvent(self, ev):
        d = ev.delta() / 120
        c, b, e = self.entered(), 'NoButton', 'Wheel'
        if not self.activated or 'pass' == self.activated.mouse[b][e](picked=c, delta=d):
            c.mouse[b][e](picked=c, delta=d)


def init_orientation_marker(i):
    """
    :param i: vtk.vtkRenderWindowInteractor
    :return: vtk.vtkOrientationMarkerWidget
    """

    vtk.vtkMapper.SetResolveCoincidentTopologyToPolygonOffset()

    cube = vtk.vtkAnnotatedCubeActor()
    cube.SetFaceTextScale(0.65)
    cube.SetXPlusFaceText("L")
    cube.SetXMinusFaceText("R")
    cube.SetYPlusFaceText("P")
    cube.SetYMinusFaceText("A")
    cube.SetZPlusFaceText("S")
    cube.SetZMinusFaceText("I")

    prop = cube.GetCubeProperty()
    prop.SetColor(0.75, 0.75, 1)

    prop = cube.GetTextEdgesProperty()
    prop.SetLineWidth(1)
    prop.SetDiffuse(0)
    prop.SetAmbient(1)
    prop.SetColor(0, 0, 0)

    prop = cube.GetXPlusFaceProperty()
    prop.SetColor(1, 0, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetXMinusFaceProperty()
    prop.SetColor(1, 0, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetYPlusFaceProperty()
    prop.SetColor(0, 1, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetYMinusFaceProperty()
    prop.SetColor(0, 1, 0)
    prop.SetInterpolationToFlat()
    prop = cube.GetZPlusFaceProperty()
    prop.SetColor(0, 0, 1)
    prop.SetInterpolationToFlat()
    prop = cube.GetZMinusFaceProperty()
    prop.SetColor(0, 0, 1)
    prop.SetInterpolationToFlat()

    axes = vtk.vtkAxesActor()

    axes.SetShaftTypeToCylinder()
    axes.SetXAxisLabelText("X")
    axes.SetYAxisLabelText("Y")
    axes.SetZAxisLabelText("Z")

    axes.SetTotalLength(1.5, 1.5, 1.5)
    axes.SetCylinderRadius(0.500 * axes.GetCylinderRadius())
    axes.SetConeRadius(1.025 * axes.GetConeRadius())
    axes.SetSphereRadius(1.500 * axes.GetSphereRadius())

    prop = axes.GetXAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(1, 0, 0)

    prop = axes.GetYAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(0, 1, 0)

    prop = axes.GetZAxisCaptionActor2D().GetCaptionTextProperty()
    prop.BoldOn()
    prop.ItalicOn()
    prop.ShadowOn()
    prop.SetFontFamilyToArial()
    prop.SetColor(0, 0, 1)

    assembly = vtk.vtkPropAssembly()
    assembly.AddPart(axes)
    assembly.AddPart(cube)

    ort = vtk.vtkOrientationMarkerWidget()
    ort.SetOrientationMarker(assembly)
    ort.SetViewport(0.0, 0.0, 0.2, 0.2)
    ort.SetInteractor(i)
    ort.SetEnabled(1)
    ort.SetInteractive(0)
    ort.On()

    return ort


if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setOrganizationName('vmi')
    app.setApplicationName('vmi')

    w = View()
    w.show()
    w._Renderer.SetBackground(1, 0.5, 0.5)

    import shelve

    with shelve.open('C:/Users/Medraw/Desktop/d') as db:
        db['w'] = w

    with shelve.open('C:/Users/Medraw/Desktop/d') as db:
        v = db['w']
        v.show()

    sys.exit(app.exec_())



from CGAL.CGAL_AABB_tree import *
from CGAL.CGAL_Kernel import *
from CGAL.CGAL_Convex_hull_2 import *
from CGAL.CGAL_Mesh_3 import *
from CGAL.CGAL_Spatial_searching import *
from CGAL.CGAL_Triangulation_2 import *
from CGAL.CGAL_HalfedgeDS import *
from CGAL.CGAL_Triangulation_3 import *
from CGAL.CGAL_Alpha_shape_2 import *
from CGAL.CGAL_Polyhedron_3 import *
from CGAL.CGAL_Voronoi_diagram_2 import *
from CGAL.CGAL_Mesh_2 import *
from CGAL.CGAL_Interpolation import *
from CGAL.CGAL_Box_intersection_d import *
from CGAL.CGAL_Convex_hull_3 import *
from CGAL.CGAL_Point_set_processing_3 import *
from CGAL.CGAL_Polyline_simplification_2 import *
from CGAL.CGAL_Advancing_front_surface_reconstruction import *
from CGAL.CGAL_Polygon_mesh_processing import *
from CGAL import *

import vmi
import vtk
import numpy as np


def cgPh_Pd(pd: vtk.vtkPolyData):
    num_vertex = pd.GetNumberOfPoints()
    num_facet = 0
    for i in range(pd.GetNumberOfCells()):
        cell: vtk.vtkCell = pd.GetCell(i)
        if cell.GetCellType() in [vtk.VTK_TRIANGLE, vtk.VTK_QUAD, vtk.VTK_POLYGON]:
            num_facet += 1

    m = Polyhedron_modifier()
    m.begin_surface(num_vertex, num_facet)

    for i in range(pd.GetNumberOfPoints()):
        pt = pd.GetPoint(i)
        m.add_vertex(Point_3(pt[0], pt[1], pt[2]))

    for cell_id in range(pd.GetNumberOfCells()):
        cell: vtk.vtkCell = pd.GetCell(cell_id)
        if cell.GetCellType() in [vtk.VTK_TRIANGLE, vtk.VTK_QUAD, vtk.VTK_POLYGON]:
            m.begin_facet()
            for ii in range(cell.GetPointIds().GetNumberOfIds()):
                pid = cell.GetPointId(ii)
                m.add_vertex_to_facet(pid)
            m.end_facet()

    m.end_surface()

    ph = Polyhedron_3()
    ph.delegate(m)
    return ph


def cgPd_Ph(ph: Polyhedron_3):
    pd = vtk.vtkPolyData()
    pd.SetPoints(vtk.vtkPoints())
    pd.SetPolys(vtk.vtkCellArray())

    points, i = {}, 0
    for vertex in ph.vertices():
        p = vertex.point()
        pd.GetPoints().InsertNextPoint([p.x(), p.y(), p.z()])
        points[str(p)] = i
        i += 1

    for facet in ph.facets():
        halfedge = facet.halfedge()
        cell_ids = []

        while len(cell_ids) == 0 or halfedge != facet.halfedge():
            cell_ids.append(points[str(halfedge.vertex().point())])
            halfedge = halfedge.next()

        n = len(cell_ids)

        if n == 3:
            cell = vtk.vtkTriangle()
            for i in range(n):
                cell.GetPointIds().SetId(i, cell_ids[i])
            pd.GetPolys().InsertNextCell(cell)
        elif n == 4:
            cell = vtk.vtkQuad()
            for i in range(n):
                cell.GetPointIds().SetId(i, cell_ids[i])
            pd.GetPolys().InsertNextCell(cell)
        elif n > 4:
            cell = vtk.vtkPolygon()
            cell.GetPointIds().SetNumberOfIds(n)
            for i in range(n):
                cell.GetPointIds().SetId(i, cell_ids[i])
            pd.GetPolys().InsertNextCell(cell)

    return pd


def pdRemesh(pd: vtk.vtkPolyData, target_edge_length=1, number_of_iterations=10):
    ph = vmi.cgPh_Pd(pd)
    isotropic_remeshing([_ for _ in ph.facets()], target_edge_length, ph, number_of_iterations)
    pd = vmi.cgPd_Ph(ph)
    return pd


def pdTriangulate_Polyline(pts):
    points = pts.copy()
    for i in range(len(points)):
        if not isinstance(points[i], Point_3):
            points[i] = Point_3(points[i][0], points[i][1], points[i][2])

    cell_ids = []
    triangulate_hole_polyline(points, cell_ids)

    pd = vtk.vtkPolyData()
    pd.SetPoints(vtk.vtkPoints())
    pd.SetPolys(vtk.vtkCellArray())

    for pt in points:
        pd.GetPoints().InsertNextPoint(pt.x(), pt.y(), pt.z())

    for cell_id in cell_ids:
        cell = vtk.vtkTriangle()
        cell.GetPointIds().SetId(0, cell_id.first)
        cell.GetPointIds().SetId(1, cell_id.second)
        cell.GetPointIds().SetId(2, cell_id.third)
        pd.GetPolys().InsertNextCell(cell)

    return pd


def cgBorder_Halfedges(ph: Polyhedron_3):
    """提取多面体ph的边界半边，仅支持单个连通多面体"""
    hs = []
    for facet in ph.facets():
        n, h = 0, facet.halfedge()

        while n == 0 or h != facet.halfedge():
            n, h = n + 1, h.next()

            if h in hs:
                hs.remove(h)
            elif h.opposite() in hs:
                hs.remove(h.opposite())
            else:
                hs.append(h)

    if len(hs) > 0:
        hs_border = [hs[0]]
        hv = {h.opposite().vertex(): h for h in hs}

        while hs_border[-1].vertex() in hv:
            h_next = hv[hs_border[-1].vertex()]
            if h_next not in hs_border:
                hs_border.append(h_next)
            else:
                break

        hs = hs_border

    return hs


def cgBorder_Pts(ph: Polyhedron_3):
    hs, pts = cgBorder_Halfedges(ph), []
    for h in hs:
        pt = h.vertex().point()
        pts.append([pt.x(), pt.y(), pt.z()])
    return pts


if __name__ == '__main__':
    points = [Point_3(0, 0, 0), Point_3(1, 0, 0), Point_3(0, 1, 0)]
    third_points = [Point_3(0.5, -1, 0), Point_3(1, 1, 0), Point_3(-1, 0.5, 0)]

    pd = pdTriangulate_Polyline(points)

    view = vmi.View()  # 视图

    actor = vmi.PolyActor(view, rgb=[1, 0, 0])
    actor.clone(pd)

    view.cameraFit()
    vmi.appexec(view)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import vtk
import vmi
import numpy as np


def pdExtract_Largest(pd: vtk.vtkPolyData):
    connectivity = vtk.vtkPolyDataConnectivityFilter()
    connectivity.SetInputData(pd)
    connectivity.SetExtractionModeToLargestRegion()
    connectivity.Update()
    return connectivity.GetOutput()


def pdExtract_Closest(pd: vtk.vtkPolyData, closest_point):
    connectivity = vtk.vtkPolyDataConnectivityFilter()
    connectivity.SetInputData(pd)
    connectivity.SetExtractionModeToClosestPointRegion()
    connectivity.SetClosestPoint([closest_point[i] for i in range(3)])
    connectivity.Update()
    return connectivity.GetOutput()


def pdSmoothWindowedSinc(pd: vtk.vtkPolyData, iter_num=20):
    smooth = vtk.vtkWindowedSincPolyDataFilter()
    smooth.SetInputData(pd)
    smooth.SetNumberOfIterations(iter_num)
    smooth.SetPassBand(0.1)
    smooth.SetNonManifoldSmoothing(1)
    smooth.SetNormalizeCoordinates(1)
    smooth.SetFeatureAngle(60)
    smooth.SetBoundarySmoothing(0)
    smooth.SetFeatureEdgeSmoothing(0)
    smooth.Update()
    return smooth.GetOutput()


def pdSmoothLaplacian(pd: vtk.vtkPolyData, iter_num=20):
    smooth = vtk.vtkSmoothPolyDataFilter()
    smooth.SetInputData(pd)
    smooth.SetNumberOfIterations(iter_num)
    smooth.SetConvergence(0)
    smooth.SetRelaxationFactor(0.33)
    smooth.SetFeatureAngle(60)
    smooth.SetBoundarySmoothing(0)
    smooth.SetFeatureEdgeSmoothing(0)
    smooth.Update()
    return smooth.GetOutput()


def pdNormals(pd: vtk.vtkPolyData):
    if pd.GetPointData().GetNormals() is None or pd.GetCellData().GetNormals() is None:
        normals = vtk.vtkPolyDataNormals()
        normals.SetInputData(pd)
        normals.SetComputePointNormals(1)
        normals.SetComputeCellNormals(1)
        normals.Update()
        return normals.GetOutput()
    return pd


def pdTriangle(pd: vtk.vtkPolyData, pass_verts=1, pass_lines=1):
    triangle = vtk.vtkTriangleFilter()
    triangle.SetPassVerts(pass_verts)
    triangle.SetPassLines(pass_lines)
    triangle.SetInputData(pd)
    triangle.Update()
    return triangle.GetOutput()


def pdClip_Implicit(pd: vtk.vtkPolyData, clip_function: vtk.vtkImplicitFunction):
    clip = vtk.vtkClipPolyData()
    clip.SetClipFunction(clip_function)
    clip.SetInputData(pdNormals(pd))
    clip.Update()
    return [clip.GetOutput(), clip.GetClippedOutput()]


def pdClip_Pd(pd: vtk.vtkPolyData, clip_pd: vtk.vtkPolyData):
    clip_function = vtk.vtkImplicitPolyDataDistance()
    clip_function.SetInput(pdNormals(clip_pd))
    return pdClip_Implicit(pd, clip_function)


def pdAppend(pds):
    append = vtk.vtkAppendPolyData()
    for pd in pds:
        append.AddInputData(pd)
    append.Update()
    return append.GetOutput()


def pdClean(pd: vtk.vtkPolyData, point_merging=1):
    clean = vtk.vtkCleanPolyData()
    clean.SetInputData(pd)
    clean.SetPointMerging(point_merging)
    clean.Update()
    return clean.GetOutput()


def pdBoolean_Union(pd0: vtk.vtkPolyData, pd1: vtk.vtkPolyData):
    boolean = vtk.vtkLoopBooleanPolyDataFilter()
    boolean.SetOperationToUnion()
    boolean.SetInputData(0, pdNormals(pd0))
    boolean.SetInputData(1, pdNormals(pd1))
    boolean.Update()
    return boolean.GetOutput()


def pdBoolean_Intersection(pd0: vtk.vtkPolyData, pd1: vtk.vtkPolyData):
    boolean = vtk.vtkLoopBooleanPolyDataFilter()
    boolean.SetOperationToIntersection()
    boolean.SetInputData(0, pdNormals(pd0))
    boolean.SetInputData(1, pdNormals(pd1))
    boolean.Update()
    return boolean.GetOutput()


def pdBoolean_Difference(pd0: vtk.vtkPolyData, pd1: vtk.vtkPolyData):
    boolean = vtk.vtkLoopBooleanPolyDataFilter()
    boolean.SetOperationToDifference()
    boolean.SetInputData(0, pdNormals(pd0))
    boolean.SetInputData(1, pdNormals(pd1))
    boolean.Update()
    return boolean.GetOutput()


def pdRayCast_Pt(pt, vt, base: vtk.vtkPolyData, base_kd=None, base_obb=None):
    pt = np.array([pt[i] for i in range(3)])
    vt = np.array([vt[i] for i in range(3)])
    vt /= np.linalg.norm(vt)

    bnd = base.GetBounds()
    d1 = np.array([bnd[1] - bnd[0], bnd[3] - bnd[2], bnd[5] - bnd[4]])
    d1 = np.linalg.norm(d1)

    if base_kd is None:
        base_kd = vtk.vtkKdTreePointLocator()
        base_kd.SetDataSet(base)
        base_kd.BuildLocator()

    cid = base_kd.FindClosestPoint(pt)

    cpt = np.array(base.GetPoint(cid))
    d2 = np.linalg.norm(pt - cpt)

    pt_last = pt + (d1 + d2) * vt

    if base_obb is None:
        base_obb = vtk.vtkOBBTree()
        base_obb.SetDataSet(base)
        base_obb.BuildLocator()

    pts = vtk.vtkPoints()
    base_obb.IntersectWithLine(pt, pt_last, pts, None)

    if pts.GetNumberOfPoints() > 0:
        return np.linalg.norm(pt - np.array(pts.GetPoint(0)))
    else:
        return None


def pdRayCast_Pd(pd: vtk.vtkPolyData, vt, base: vtk.vtkPolyData, default_origin):
    vt = np.array([vt[i] for i in range(3)])
    vt /= np.linalg.norm(vt)

    base_kd = vtk.vtkKdTreePointLocator()
    base_kd.SetDataSet(base)
    base_kd.BuildLocator()

    base_obb = vtk.vtkOBBTree()
    base_obb.SetDataSet(base)
    base_obb.BuildLocator()

    out = vtk.vtkPolyData()
    out.DeepCopy(pd)

    for i in range(pd.GetNumberOfPoints()):
        pt = pd.GetPoint(i)
        d = pdRayCast_Pt(pt, vt, base, base_kd, base_obb)

        if d is None:
            pt = vmi.ptOnPlane(pt, default_origin, vt)
        else:
            pt = pt + d * vt
        out.GetPoints().SetPoint(i, pt)

    return out


def pdOBB_Pts(pts):
    if isinstance(pts, vtk.vtkPolyData):
        pts = pts.GetPoints()
    elif not isinstance(pts, vtk.vtkPoints):
        points = vtk.vtkPoints()
        for pt in pts:
            points.InsertNextPoint([pt[0], pt[1], pt[2]])
        pts = points

    if pts.GetNumberOfPoints() > 0:
        args = {'origin': [0, 0, 0], 'axis': [[0, 0, 0], [0, 0, 0], [0, 0, 0]], 'size': [0, 0, 0]}
        vtk.vtkOBBTree().ComputeOBB(pts, args['origin'], args['axis'][0], args['axis'][1], args['axis'][2],
                                    args['size'])

        args['center'] = args['origin']
        for i in args:
            args[i] = np.array(args[i])

        for j in range(3):
            args['axis'][j] = args['axis'][j] / np.linalg.norm(args['axis'][j])
            args['center'] = [args['center'][i] + 0.5 * args['size'][j] * args['axis'][j][i] for i in range(3)]
        return args


def pdOffset_Normals(pd: vtk.vtkPolyData, d):
    pd = pdNormals(pd)
    pd_out = vtk.vtkPolyData()
    pd_out.DeepCopy(pd)

    normals = pd.GetPointData().GetNormals()
    for i in range(pd.GetNumberOfPoints()):
        n = [0, 0, 0]
        normals.GetTypedTuple(i, n)
        n = np.array(n)
        n /= np.linalg.norm(n)

        pt = np.array(pd.GetPoint(i)) + d*n
        pd_out.GetPoints().SetPoint(i, pt)
    return pd_out


def pdMatrix(pd: vtk.vtkPolyData, matrix: vtk.vtkMatrix4x4):
    t = vtk.vtkTransform()
    t.SetMatrix(matrix)
    return pdTransform(pd, t)


def pdTransform(pd: vtk.vtkPolyData, t: vtk.vtkTransform):
    tf = vtk.vtkTransformPolyDataFilter()
    tf.SetTransform(t)
    tf.SetInputData(pd)
    tf.Update()
    return tf.GetOutput()


def pdBorder_Pts(pd: vtk.vtkPolyData):
    ph = vmi.cgPh_Pd(pd)
    return vmi.cgBorder_Pts(ph)



import vmi


def read_dicom():
    dcmdir = vmi.askDirectory()  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                return series.read()  # 读取DICOM系列为图像数据


if __name__ == '__main__':
    image = read_dicom()  # 读取DICOM路径
    if image is None:
        vmi.appexit()

    view = vmi.View()  # 视图

    image_slice = vmi.ImageSlice(view)  # 断层显示
    image_slice.clone(image)  # 载入读取到的DICOM图像
    image_slice.slicePlane(face='axial', origin='center')  # 设置断层图像显示横断面，位置居中
    view.cameraFit()  # 自动调整视图的视野范围

    vmi.appexec(view)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import vmi

if __name__ == '__main__':
    from vmi.example.read_dicom_slice_view import read_dicom
    from vmi.example.surface_volume_rendering import view_surface

    image = read_dicom()  # 读取DICOM路径
    if image is None:
        vmi.appexit()

    iso_value = vmi.askInt(-1000, 200, 3000, None, 'HU')  # 用户输入-1000HU至3000HU范围内的一个CT值
    if iso_value is None:
        vmi.appexit()

    pts = []  # 取点列表


    def rebuild():
        """构造曲线"""
        if len(pts) > 1:  # 如果有2个点以上，可以构造曲线
            curve = vmi.ccBSpline_Kochanek(pts)  # 由多个点坐标构造几何曲线
            edge = vmi.ccEdge(curve)  # 由几何曲线构造拓扑边
            prop_curve.clone(edge)  # 传入shape数据
        else:  # 如果没有点或只有1个点，不能构造曲线
            prop_curve.clone()  # 重置shape数据


    def LeftButtonPressRelease(**kwargs):
        """响应左键点击"""
        if kwargs['picked'] is prop_mesh:  # 判断鼠标拾取到的是模型
            if kwargs['double'] is False:  # 鼠标单击时
                pt = view.mouseOnPropCell()  # 在视图中取物体表面点
                pts.append(pt)  # 添加进取点列表
                rebuild()  # 重新构造曲线
            else:  # 鼠标双击时
                pts.clear()  # 清空取点列表
                rebuild()  # 重新构造曲线
            return prop_mesh.mouseRelease()


    view = vmi.View()  # 视图

    prop_mesh = view_surface(view, image, iso_value)  # 三维重建获得单元数据和表示
    prop_mesh.color([1, 1, 0.6])  # 颜色
    prop_mesh.pickable(True)  # 设置可拾取
    prop_mesh.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease  # 连接鼠标左键点击

    prop_curve = vmi.PolyActor(view)  # 曲线表示
    prop_curve.color([1, 0.4, 0.4])  # 颜色
    prop_curve.size(line=3)  # 设置线宽3个屏幕像素

    vmi.appexec(view)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import vmi
import numpy as np


def reslice_blend():
    left_top = view[1].pickFPlane([0, 0])  # 拾取视图左上角的焦平面点
    right_top = view[1].pickFPlane([view[1].width(), 0])  # 拾取视图右上角的焦平面点
    left_bottom = view[1].pickFPlane([0, view[1].height()])  # 拾取视图左下角的焦平面点

    cs = view[1].cameraCS()  # 获得视角坐标系

    xl = np.linalg.norm(left_top - right_top)  # 视图左右方向的真实尺度
    yl = np.linalg.norm(left_top - left_bottom)  # 视图上下方向的真实尺度
    zl = vmi.imSize_Vt(image, cs.axis(2))  # 图像沿视图入射方向的最大包围盒尺寸

    pt = vmi.imCenter(image) - 0.5 * zl * cs.axis(2)  # 计算图像包围盒中心到近平面投影点
    cs.origin(vmi.ptOnPlane(cs.origin(), pt, cs.axis(2)))  # 将坐标系原点投影到近平面

    # 重采样混叠，背景体素值-1024，平均密度投影，忽略-200以下的体素值
    image_blend = vmi.imReslice_Blend(image, cs, [xl, yl, zl], -1024, 'mean', -200)

    # 更新视图
    blend_prop.clone(image_blend)
    blend_prop.windowSoft()
    view[0].cameraFit()


if __name__ == '__main__':
    from PySide2.QtWidgets import QWidget, QHBoxLayout
    from vmi.example.read_dicom_slice_view import read_dicom

    image = read_dicom()  # 读取DICOM路径
    if image is None:
        vmi.appexit()

    iso_value = vmi.askInt(-1000, 200, 3000, None, 'HU')  # 用户输入-1000HU至3000HU范围内的一个CT值
    if iso_value is None:
        vmi.appexit()

    view = [vmi.View(), vmi.View()]  # 左右视图
    blend_prop = vmi.ImageSlice(view[0])  # 左视图显示混叠图像

    menu = vmi.Menu()  # 右键菜单
    view[1].mouse['RightButton']['PressRelease'] = menu.menuexec  # 连接右视图的右键信号到菜单

    menu.menu.addSection('自定义')  # 添加分割线
    menu.menu.addAction('重采样混叠').triggered.connect(reslice_blend)  # 创建并连接菜单选项
    menu.menu.addSection('视图')  # 添加分割线
    menu.menu.addActions(view[1].menu.actions())  # 添加右视图原有菜单

    mesh_prop = vmi.PolyActor(view[1], rgb=[1, 1, 0.6])  # 右视图显示模型
    mesh_prop.clone(vmi.imIsosurface(image, iso_value))  # 三维重建并显示
    view[1].cameraCoronal()
    view[1].cameraFit()  # 重置视野范围

    # 创建窗口包含两个视图
    widget = QWidget()
    layout = QHBoxLayout(widget)
    layout.addWidget(view[0])
    layout.addWidget(view[1])

    vmi.appexec(widget)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import sys
from typing import List, Dict, Optional, Union

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
import vmi

_SizePolicyMinimum = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
tr = QObject()
tr = tr.tr


def askOpenFile(nameFilter: str = '*',
                title: str = tr('请选择打开文件 (Please select open file)')) -> Optional[str]:
    """
    打开文件对话框

    :param nameFilter: 对话框内显示的文件名
    :param title: 对话框名称
    :return: str - 有效选择的文件名，None - 没有有效选择
    """
    s = QSettings()

    d = QFileDialog(None, title, s.value('Last' + title, defaultValue=vmi.desktopPath))
    d.setAcceptMode(QFileDialog.AcceptOpen)
    d.setFileMode(QFileDialog.ExistingFile)
    d.setViewMode(QFileDialog.Detail)
    d.setFilter(QDir.AllEntries | QDir.NoSymLinks | QDir.NoDotAndDotDot)
    d.setNameFilter(nameFilter)

    if d.exec_() != QDialog.Accepted or len(d.selectedFiles()) < 1:
        return None

    s.setValue('Last' + title, d.selectedFiles()[0])
    return d.selectedFiles()[0]


def askSaveFile(suffix: str = '', nameFilter: str = '*',
                title: str = tr('请选择保存文件 (Please select save file)')) -> Optional[str]:
    """
    保存文件对话框

    :param suffix: 默认后缀名
    :param nameFilter: 对话框内显示的文件名
    :param title: 对话框名称
    :return: str - 有效选择的文件名，None - 没有有效选择
    """
    s = QSettings()

    d = QFileDialog(None, title, s.value('Last' + title, defaultValue=vmi.desktopPath))
    d.setAcceptMode(QFileDialog.AcceptSave)
    d.setFileMode(QFileDialog.AnyFile)
    d.setViewMode(QFileDialog.Detail)
    d.setFilter(QDir.AllEntries | QDir.NoSymLinks | QDir.NoDotAndDotDot)
    d.setDefaultSuffix(suffix)
    d.setNameFilter(nameFilter)

    if nameFilter is None:
        d.setNameFilter(suffix)

    if d.exec_() != QDialog.Accepted or len(d.selectedFiles()) < 1:
        return None

    s.setValue('Last' + title, d.selectedFiles()[0])
    return d.selectedFiles()[0]


def askOpenFiles(nameFilter: str = '*',
                 title: str = tr('请选择打开文件 (Please select open files)')) -> Optional[List[str]]:
    """
    打开多个文件对话框

    :param nameFilter: 对话框内显示的文件名
    :param title: 对话框名称
    :return: List[str] - 有效选择的多个文件名，None - 没有有效选择
    """
    s = QSettings()

    d = QFileDialog(None, title, s.value('Last' + title, defaultValue=vmi.desktopPath))
    d.setAcceptMode(QFileDialog.AcceptOpen)
    d.setFileMode(QFileDialog.ExistingFiles)
    d.setViewMode(QFileDialog.Detail)
    d.setFilter(QDir.AllEntries | QDir.NoSymLinks | QDir.NoDotAndDotDot)
    d.setNameFilter(nameFilter)

    if d.exec_() != QDialog.Accepted or len(d.selectedFiles()) < 1:
        return None

    s.setValue('Last' + title, QFileInfo(d.selectedFiles()[0]).absolutePath())
    return d.selectedFiles()


def askDirectory(title: str = tr('请选择文件夹 (Please select directory)')) -> Optional[str]:
    """
    打开文件目录对话框

    :param title: 对话框名称
    :return: str - 有效选择的文件目录，None - 没有有效选择
    """
    s = QSettings()

    d = QFileDialog(None, title, s.value('Last' + title, defaultValue=vmi.desktopPath))
    d.setAcceptMode(QFileDialog.AcceptOpen)
    d.setFileMode(QFileDialog.Directory)
    d.setViewMode(QFileDialog.Detail)
    d.setFilter(QDir.AllEntries | QDir.NoSymLinks | QDir.NoDotAndDotDot)

    if d.exec_() != QDialog.Accepted or len(d.selectedFiles()) < 1:
        return None

    s.setValue('Last' + title, d.selectedFiles()[0])
    return d.selectedFiles()[0]


def _dialog_ok_cancel(widgets, title) -> QDialog:
    dialog = QDialog()
    dialog.setSizePolicy(_SizePolicyMinimum)
    dialog.setWindowTitle(title)
    dialog.setLayout(QVBoxLayout())

    button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
    button.setSizePolicy(_SizePolicyMinimum)
    button.accepted.connect(dialog.accept)
    button.rejected.connect(dialog.reject)

    for w in widgets:
        w.setParent(dialog)
        dialog.layout().addWidget(w)
    dialog.layout().addWidget(button)
    return dialog


def askSeries(series: List[vmi.SeriesData],
              title: str = tr('请选择目标系列 (Please select the target series)')) -> Optional[vmi.SeriesData]:
    """
    选择DICOM系列

    :param series: 候选的DICOM系列列表
    :param title: 对话框名称
    :return: vmi.SeriesData - 有效选择的DICOM系列，None - 没有有效选择
    """
    items = [str(i) for i in series]

    def tabCloseRequested(i):
        del items[i]
        tabs.removeTab(i)

    tabs = QTabWidget()
    tabs.setTabsClosable(True)
    tabs.setSizePolicy(_SizePolicyMinimum)
    tabs.tabCloseRequested.connect(tabCloseRequested)

    for i in items:
        label = QLabel(text=i)
        tabs.addTab(label, str(tabs.count() + 1))

    d = _dialog_ok_cancel([tabs], title)
    if d.exec_() == QDialog.Accepted:
        if 0 <= tabs.currentIndex() < len(series):
            return series[tabs.currentIndex()]
    return None


def askInt(value_min: int, value: int, value_max: int, step: Optional[int] = None, prefix: str = None,
           suffix: str = None,
           title: str = tr('输入数值 (Input value)')) -> Optional[int]:
    """
    选择整形数值

    :param value_min: 最小值
    :param value: 默认值
    :param value_max: 最大值
    :param step: 最小单位
    :param prefix: 前缀
    :param suffix: 后缀
    :param title: 对话框名称
    :return: int - 有效选择的数值，None - 没有有效选择
    """
    w = QGroupBox(prefix)
    w.setSizePolicy(_SizePolicyMinimum)
    w.setLayout(QHBoxLayout())

    spinbox = QSpinBox(w)
    spinbox.setSizePolicy(_SizePolicyMinimum)
    spinbox.setAlignment(Qt.AlignCenter)
    spinbox.setRange(value_min, value_max)
    spinbox.setValue(value)
    w.layout().addWidget(spinbox)

    if suffix is not None:
        spinbox.setSuffix(' ' + suffix)

    if step is not None:
        spinbox.setSingleStep(step)
        spinbox.setStepType(QSpinBox.DefaultStepType)
    else:
        spinbox.setStepType(QSpinBox.AdaptiveDecimalStepType)

    d = _dialog_ok_cancel([w], title)
    if d.exec_() == QDialog.Accepted:
        return spinbox.value()
    return None


def askInts(value_min: List[int], value: List[int], value_max: List[int], step: Optional[List[int]] = None,
            prefix=None, suffix=None, title=tr('输入数值 (Input value)')) -> Optional[List[int]]:
    """
    选择多个整形数值

    :param value_min: 最小值列表
    :param value: 默认值列表
    :param value_max: 最大值列表
    :param step: 最小单位列表
    :param prefix: 前缀
    :param suffix: 后缀
    :param title: 对话框名称
    :return: List[int] - 有效选择的多个数值，None - 没有有效选择
    """
    w = QGroupBox(prefix)
    w.setSizePolicy(_SizePolicyMinimum)
    w.setLayout(QHBoxLayout())

    spinbox = []
    for i in range(len(value)):
        spinbox.append(QSpinBox(w))
        spinbox[-1].setSizePolicy(_SizePolicyMinimum)
        spinbox[-1].setAlignment(Qt.AlignCenter)
        spinbox[-1].setRange(value_min[i], value_max[i])
        spinbox[-1].setValue(value[i])
        w.layout().addWidget(spinbox[-1])

        if suffix is not None:
            spinbox[-1].setSuffix(' ' + suffix[i])

        if step is not None:
            spinbox[-1].setSingleStep(step[i])
            spinbox[-1].setStepType(QSpinBox.DefaultStepType)
        else:
            spinbox[-1].setStepType(QSpinBox.AdaptiveDecimalStepType)

    d = _dialog_ok_cancel([w], title)
    if d.exec_() == QDialog.Accepted:
        return [spinbox[i].value() for i in range(len(value))]
    return None


def askFloat(value_min: float, value: float, value_max: float, decimals: int, step: Optional[float] = None,
             prefix=None, suffix=None, title: str = tr('输入数值 (Input value)')) -> Optional[float]:
    """
    选择浮点数值

    :param value_min: 最小值列表
    :param value: 默认值列表
    :param value_max: 最大值列表
    :param decimals: 小数位数
    :param step: 最小单位列表
    :param prefix: 前缀
    :param suffix: 后缀
    :param title: 对话框名称
    :return: float - 有效选择的数值，None - 没有有效选择
    """
    w = QGroupBox(prefix)
    w.setSizePolicy(_SizePolicyMinimum)
    w.setLayout(QHBoxLayout())

    spinbox = QDoubleSpinBox(w)
    spinbox.setSizePolicy(_SizePolicyMinimum)
    spinbox.setAlignment(Qt.AlignCenter)
    spinbox.setDecimals(decimals)
    spinbox.setRange(value_min, value_max)
    spinbox.setValue(value)
    w.layout().addWidget(spinbox)

    if suffix is not None:
        spinbox.setSuffix(' ' + suffix)

    if step is not None:
        spinbox.setSingleStep(step)
        spinbox.setStepType(QSpinBox.DefaultStepType)
    else:
        spinbox.setStepType(QSpinBox.AdaptiveDecimalStepType)

    d = _dialog_ok_cancel([w], title)
    if d.exec_() == QDialog.Accepted:
        return spinbox.value()
    return None


def askFloats(value_min: List[float], value: List[float], value_max: List[float], decimals: int,
              step: Optional[float] = None, prefix=None, suffix=None,
              title: str = tr('输入数值 (Input value)')) -> Optional[List[float]]:
    """
    选择多个浮点数值

    :param value_min: 最小值列表
    :param value: 默认值列表
    :param value_max: 最大值列表
    :param decimals: 小数位数
    :param step: 最小单位列表
    :param prefix: 前缀
    :param suffix: 后缀
    :param title: 对话框名称
    :return: float - 有效选择的数值，None - 没有有效选择
    """
    w = QGroupBox(prefix)
    w.setSizePolicy(_SizePolicyMinimum)
    w.setLayout(QHBoxLayout())

    spinbox = []
    for i in range(len(value)):
        spinbox.append(QDoubleSpinBox(w))
        spinbox[-1].setSizePolicy(_SizePolicyMinimum)
        spinbox[-1].setAlignment(Qt.AlignCenter)
        spinbox[-1].setDecimals(decimals)
        spinbox[-1].setRange(value_min[i], value_max[i])
        spinbox[-1].setValue(value[i])
        w.layout().addWidget(spinbox[-1])

        if suffix is not None:
            spinbox[-1].setSuffix(' ' + suffix[i])

        if step is not None:
            spinbox[-1].setSingleStep(step[i])
            spinbox[-1].setStepType(QSpinBox.DefaultStepType)
        else:
            spinbox[-1].setStepType(QSpinBox.AdaptiveDecimalStepType)

    d = _dialog_ok_cancel([w], title)
    if d.exec_() == QDialog.Accepted:
        return [spinbox[i].value() for i in range(len(value))]
    return None


def askInfo(text: str, title: str = tr('提示 (info)')) -> None:
    """
    提示对话框

    :param text: 信息文本
    :param title: 对话框名称
    :return: None
    """
    QMessageBox.information(None, title, text)


def askYesNo(text: str, title: str = tr('是否 (yes or no)')) -> bool:
    """
    询问对话框
    :param text: 信息文本
    :param title: 对话框名称
    :return: True - 选择了是，False - 选择了否
    """
    return QMessageBox.question(None, title, text) == QMessageBox.Yes


if __name__ == '__main__':
    vmi.app.setApplicationName('vmi')

    askInt(-3071, -1, 1024, prefix='CT: [-3071, 1024]', suffix='HU')



name = 'vmi'
version = '0.3.3'

from .vrc import *
from .vapp import *
from .view import *
from .vicom import *
from .vickle import *
from .vath import *
from .vimage import *
from .vpoly import *
from .vprop import *
# from .vproptool import *
from .vcc import *
from .vcg import *
from .vask import *



# from OCC.Core.AIS import *
# from OCC.Core.Adaptor2d import *
# from OCC.Core.Adaptor3d import *
# from OCC.Core.Addons import *
# from OCC.Core.AdvApp2Var import *
# from OCC.Core.AdvApprox import *
# from OCC.Core.AppBlend import *
# from OCC.Core.AppCont import *
# from OCC.Core.AppDef import *
# from OCC.Core.AppParCurves import *
# from OCC.Core.AppStd import *
# from OCC.Core.AppStdL import *
# from OCC.Core.Approx import *
# from OCC.Core.ApproxInt import *
# from OCC.Core.Aspect import *
# from OCC.Core.BOPAlgo import *
# from OCC.Core.BOPCol import *
# from OCC.Core.BOPDS import *
# from OCC.Core.BOPInt import *
# from OCC.Core.BOPTools import *
from OCC.Core.BRep import *
from OCC.Core.BRepAdaptor import *
# from OCC.Core.BRepAlgo import *
from OCC.Core.BRepAlgoAPI import *
# from OCC.Core.BRepApprox import *
# from OCC.Core.BRepBlend import *
from OCC.Core.BRepBndLib import *
from OCC.Core.BRepBuilderAPI import *
# from OCC.Core.BRepCheck import *
# from OCC.Core.BRepClass import *
from OCC.Core.BRepClass3d import *
# from OCC.Core.BRepExtrema import *
# from OCC.Core.BRepFeat import *
# from OCC.Core.BRepFill import *
# from OCC.Core.BRepFilletAPI import *
from OCC.Core.BRepGProp import *
# from OCC.Core.BRepIntCurveSurface import *
# from OCC.Core.BRepLProp import *
# from OCC.Core.BRepLib import *
# from OCC.Core.BRepMAT2d import *
from OCC.Core.BRepMesh import *
# from OCC.Core.BRepOffset import *
from OCC.Core.BRepOffsetAPI import *
# from OCC.Core.BRepPrim import *
from OCC.Core.BRepPrimAPI import *
# from OCC.Core.BRepProj import *
from OCC.Core.BRepSweep import *
from OCC.Core.BRepTools import *
# from OCC.Core.BRepTopAdaptor import *
# from OCC.Core.BSplCLib import *
# from OCC.Core.BSplSLib import *
# from OCC.Core.BiTgte import *
# from OCC.Core.Bisector import *
# from OCC.Core.Blend import *
# from OCC.Core.BlendFunc import *
from OCC.Core.Bnd import *
# from OCC.Core.BndLib import *
# from OCC.Core.CDF import *
# from OCC.Core.CDM import *
# from OCC.Core.CPnts import *
# from OCC.Core.CSLib import *
# from OCC.Core.ChFi2d import *
# from OCC.Core.ChFi3d import *
# from OCC.Core.ChFiDS import *
# from OCC.Core.ChFiKPart import *
# from OCC.Core.Contap import *
# from OCC.Core.Convert import *
# from OCC.Core.Dico import *
# from OCC.Core.Draft import *
# from OCC.Core.DsgPrs import *
# from OCC.Core.Dynamic import *
# from OCC.Core.ElCLib import *
# from OCC.Core.ElSLib import *
# from OCC.Core.Expr import *
# from OCC.Core.ExprIntrp import *
# from OCC.Core.Extrema import *
# from OCC.Core.FEmTool import *
# from OCC.Core.FSD import *
# from OCC.Core.FairCurve import *
# from OCC.Core.FilletSurf import *
from OCC.Core.GC import *
# from OCC.Core.GCE2d import *
from OCC.Core.GCPnts import *
# from OCC.Core.GEOMAlgo import *
from OCC.Core.GProp import *
# from OCC.Core.GccAna import *
# from OCC.Core.GccEnt import *
# from OCC.Core.GccGeo import *
# from OCC.Core.GccInt import *
# from OCC.Core.GccIter import *
# from OCC.Core.Geom import *
# from OCC.Core.Geom2d import *
# from OCC.Core.Geom2dAPI import *
# from OCC.Core.Geom2dAdaptor import *
# from OCC.Core.Geom2dConvert import *
# from OCC.Core.Geom2dGcc import *
# from OCC.Core.Geom2dHatch import *
# from OCC.Core.Geom2dInt import *
# from OCC.Core.Geom2dLProp import *
from OCC.Core.GeomAPI import *
# from OCC.Core.GeomAbs import *
# from OCC.Core.GeomAdaptor import *
# from OCC.Core.GeomConvert import *
# from OCC.Core.GeomFill import *
# from OCC.Core.GeomInt import *
# from OCC.Core.GeomLProp import *
# from OCC.Core.GeomLib import *
# from OCC.Core.GeomPlate import *
from OCC.Core.GeomProjLib import *
# from OCC.Core.GeomToStep import *
# from OCC.Core.GeomTools import *
# from OCC.Core.GraphDS import *
# from OCC.Core.GraphTools import *
# from OCC.Core.Graphic3d import *
# from OCC.Core.HLRAlgo import *
# from OCC.Core.HLRAppli import *
# from OCC.Core.HLRBRep import *
# from OCC.Core.HLRTopoBRep import *
# from OCC.Core.Hatch import *
# from OCC.Core.HatchGen import *
# from OCC.Core.Hermit import *
from OCC.Core.IFSelect import *
# from OCC.Core.IGESCAFControl import *
from OCC.Core.IGESControl import *
# from OCC.Core.Image import *
# from OCC.Core.IncludeLibrary import *
# from OCC.Core.IntAna import *
# from OCC.Core.IntAna2d import *
# from OCC.Core.IntCurve import *
# from OCC.Core.IntCurveSurface import *
# from OCC.Core.IntCurvesFace import *
# from OCC.Core.IntImp import *
# from OCC.Core.IntImpParGen import *
# from OCC.Core.IntPatch import *
# from OCC.Core.IntPoly import *
# from OCC.Core.IntPolyh import *
# from OCC.Core.IntRes2d import *
# from OCC.Core.IntStart import *
# from OCC.Core.IntSurf import *
# from OCC.Core.IntTools import *
# from OCC.Core.IntWalk import *
# from OCC.Core.Interface import *
# from OCC.Core.InterfaceGraphic import *
# from OCC.Core.Intf import *
# from OCC.Core.Intrv import *
# from OCC.Core.LProp import *
# from OCC.Core.LProp3d import *
# from OCC.Core.Law import *
# from OCC.Core.LocOpe import *
# from OCC.Core.LocalAnalysis import *
# from OCC.Core.MAT import *
# from OCC.Core.MAT2d import *
# from OCC.Core.MMgt import *
# from OCC.Core.Materials import *
# from OCC.Core.MeshVS import *
# from OCC.Core.Message import *
# from OCC.Core.NCollection import *
# from OCC.Core.NIS import *
# from OCC.Core.NLPlate import *
# from OCC.Core.OSD import *
# from OCC.Core.PCDM import *
# from OCC.Core.PLib import *
# from OCC.Core.Plate import *
# from OCC.Core.Plugin import *
from OCC.Core.Poly import *
from OCC.Core.Precision import *
# from OCC.Core.Primitives import *
# from OCC.Core.ProjLib import *
# from OCC.Core.Prs3d import *
# from OCC.Core.PrsMgr import *
# from OCC.Core.Quantity import *
# from OCC.Core.RWStepAP203 import *
# from OCC.Core.RWStepAP214 import *
# from OCC.Core.RWStepBasic import *
# from OCC.Core.RWStepDimTol import *
# from OCC.Core.RWStepElement import *
# from OCC.Core.RWStepFEA import *
# from OCC.Core.RWStepGeom import *
# from OCC.Core.RWStepRepr import *
# from OCC.Core.RWStepShape import *
# from OCC.Core.RWStepVisual import *
# from OCC.Core.RWStl import *
# from OCC.Core.Resource import *
# from OCC.Core.STEPCAFControl import *
# from OCC.Core.STEPConstruct import *
from OCC.Core.STEPControl import *
# from OCC.Core.STEPEdit import *
# from OCC.Core.STEPSelections import *
# from OCC.Core.Select3D import *
# from OCC.Core.SelectBasics import *
# from OCC.Core.SelectMgr import *
# from OCC.Core.ShapeAlgo import *
# from OCC.Core.ShapeAnalysis import *
# from OCC.Core.ShapeBuild import *
# from OCC.Core.ShapeConstruct import *
# from OCC.Core.ShapeCustom import *
# from OCC.Core.ShapeExtend import *
# from OCC.Core.ShapeFix import *
# from OCC.Core.ShapeProcess import *
# from OCC.Core.ShapeProcessAPI import *
# from OCC.Core.ShapeUpgrade import *
# from OCC.Core.SortTools import *
# from OCC.Core.Standard import *
# from OCC.Core.StdFail import *
# from OCC.Core.StdPrs import *
# from OCC.Core.StdSelect import *
# from OCC.Core.StepAP203 import *
# from OCC.Core.StepAP209 import *
# from OCC.Core.StepAP214 import *
# from OCC.Core.StepBasic import *
# from OCC.Core.StepDimTol import *
# from OCC.Core.StepElement import *
# from OCC.Core.StepFEA import *
# from OCC.Core.StepGeom import *
# from OCC.Core.StepRepr import *
# from OCC.Core.StepShape import *
# from OCC.Core.StepToGeom import *
# from OCC.Core.StepToTopoDS import *
# from OCC.Core.StepVisual import *
from OCC.Core.StlAPI import *
from OCC.Core.StlMesh import *
# from OCC.Core.StlTransfer import *
# from OCC.Core.Storage import *
# from OCC.Core.Sweep import *
# from OCC.Core.TColGeom import *
# from OCC.Core.TColGeom2d import *
# from OCC.Core.TColQuantity import *
from OCC.Core.TColStd import *
from OCC.Core.TColgp import *
# from OCC.Core.TCollection import *
# from OCC.Core.TDF import *
# from OCC.Core.TDataStd import *
# from OCC.Core.TDataXtd import *
# from OCC.Core.TDocStd import *
# from OCC.Core.TFunction import *
# from OCC.Core.TNaming import *
# from OCC.Core.TPrsStd import *
# from OCC.Core.TShort import *
from OCC.Core.TopAbs import *
# from OCC.Core.TopBas import *
# from OCC.Core.TopClass import *
# from OCC.Core.TopCnx import *
from OCC.Core.TopExp import *
from OCC.Core.TopLoc import *
# from OCC.Core.TopOpeBRep import *
# from OCC.Core.TopOpeBRepBuild import *
# from OCC.Core.TopOpeBRepDS import *
# from OCC.Core.TopOpeBRepTool import *
from OCC.Core.TopTools import *
# from OCC.Core.TopTrans import *
from OCC.Core.TopoDS import *
# from OCC.Core.TopoDSToStep import *
# from OCC.Core.Transfer import *
# from OCC.Core.TransferBRep import *
# from OCC.Core.UTL import *
# from OCC.Core.Units import *
# from OCC.Core.UnitsAPI import *
# from OCC.Core.V3d import *
# from OCC.Core.Visual3d import *
# from OCC.Core.Visualization import *
# from OCC.Core.Voxel import *
# from OCC.Core.XBRepMesh import *
# from OCC.Core.XCAFApp import *
# from OCC.Core.XCAFDoc import *
# from OCC.Core.XCAFPrs import *
# from OCC.Core.XSControl import *
# from OCC.Core.gce import *
from OCC.Core.gp import *
# from OCC.Core.math import *

import tempfile
import pathlib
from PySide2.QtCore import *
from PySide2.QtWidgets import *
import vtk
import vmi
import numpy as np

import warnings

from typing import List


def ignore_warnings(*args):
    pass


warnings.simplefilter("ignore")
warnings.simplefilter = ignore_warnings

tr = QObject()
tr = tr.tr

ccType = [TopAbs_COMPOUND,
          TopAbs_COMPSOLID,
          TopAbs_SOLID,
          TopAbs_SHELL,
          TopAbs_FACE,
          TopAbs_WIRE,
          TopAbs_EDGE,
          TopAbs_VERTEX,
          TopAbs_SHAPE]

ccTypeCast = {TopAbs_COMPOUND: topods.Compound,
              TopAbs_COMPSOLID: topods.CompSolid,
              TopAbs_SOLID: topods.Solid,
              TopAbs_SHELL: topods.Shell,
              TopAbs_FACE: topods.Face,
              TopAbs_WIRE: topods.Wire,
              TopAbs_EDGE: topods.Edge,
              TopAbs_VERTEX: topods.Vertex}


def ccExplore_Recursive(sh: TopoDS_Shape, target_type: int = ccType[-1]):
    shs, exp = [], TopExp_Explorer(sh, target_type)
    while exp.More():
        shs.append(exp.Current())
        exp.Next()
    return shs


def ccExplore_Once(sh: TopoDS_Shape):
    shs, it = [], TopoDS_Iterator(sh)
    while it.More():
        shs.append(it.Value())
        it.Next()
    return shs


def ccOpenFile_IGES(file: str = None):
    if file is None:
        file = vmi.askOpenFile(tr('导入 (Import) IGES'), '*.igs')
        if file is None:
            return
    r = IGESControl_Reader()
    if r.ReadFile(file):
        r.TransferRoots()
        sh: TopoDS_Shape = r.OneShape()
        sh_sub = ccExplore_Once(sh)
        if sh.ShapeType() is TopAbs_COMPOUND and len(sh_sub) == 1:
            sh = sh_sub[0]
        return sh


def ccOpenFile_STEP(file: str = None):
    if file is None:
        file = vmi.askOpenFile(tr('导入 (Import) STEP'), '*.stp')
        if file is None:
            return
    r = STEPControl_Reader()
    if IFSelect_RetDone == r.ReadFile(file):
        r.TransferRoots()
        sh: TopoDS_Shape = r.OneShape()
        sh_sub = ccExplore_Once(sh)
        if sh.ShapeType() is TopAbs_COMPOUND and len(sh_sub) == 1:
            sh = sh_sub[0]
        return sh


def ccSaveFile_IGES(sh: TopoDS_Shape, file=None):
    if file is None:
        file = vmi.askSaveFile(tr('导出 (Export) IGES'), '*.igs')
        if file is None:
            return
    w = IGESControl_Writer()
    w.AddShape(sh)
    return w.Write(file)


def ccSaveFile_STEP(sh: TopoDS_Shape, file=None):
    if file is None:
        file = vmi.askSaveFile(tr('导出 (Export) STEP'), '*.stp')
        if file is None:
            return
    w = STEPControl_Writer()
    w.Transfer(sh, STEPControl_AsIs)
    status = w.Write(file)
    if status == IFSelect_RetDone:
        return True


def ccSaveFile_STL(sh: TopoDS_Shape, file=None):
    if file is None:
        file = vmi.askSaveFile(tr('导出 (Export) STL'), '*.stl')
        if file is None:
            return
    return stlapi.Write(sh, file, False)


def ccPd_Sh(sh: TopoDS_Shape):
    if sh.IsNull():
        return vtk.vtkPolyData()

    pts = vtk.vtkPoints()
    polydata = vtk.vtkPolyData()
    polydata.SetPoints(pts)
    polydata.Allocate()

    # 1. 离散化
    breptools.Clean(sh)

    discrete = BRepMesh_IncrementalMesh(sh, 1e-3, True, 0.5, True)
    discrete.Perform()

    # 2. 转换Vertex
    exp = TopExp_Explorer(sh, TopAbs_VERTEX)
    while exp.More():
        vertex = topods.Vertex(exp.Current())
        exp.Next()

        pnt = BRep_Tool.Pnt(vertex)
        id = polydata.GetPoints().InsertNextPoint(pnt.X(), pnt.Y(), pnt.Z())
        polydata.InsertNextCell(vtk.VTK_VERTEX, 1, [id])

    # 3. 转换Edge
    exp = TopExp_Explorer(sh, TopAbs_EDGE)
    while exp.More():
        edge = topods.Edge(exp.Current())
        exp.Next()

        if edge.IsNull() or BRep_Tool.Degenerated(edge):
            continue

        p_on_tri = Handle_Poly_PolygonOnTriangulation()
        triangulation = Handle_Poly_Triangulation()
        loc = TopLoc_Location()
        BRep_Tool.PolygonOnTriangulation(edge, p_on_tri, triangulation, loc, 1)
        p_3d = Handle_Poly_Polygon3D()

        if p_on_tri.IsNull():
            p_3d = BRep_Tool.Polygon3D(None, edge, loc)

        if p_3d.IsNull() and p_on_tri.IsNull():
            continue

        trsf = loc.Transformation()

        if not p_3d.IsNull():
            n = p_3d.NbNodes()
            pnts = p_3d.Nodes()
            ids = list(range(1, n + 1))
        else:
            n = p_on_tri.NbNodes()
            pnts = triangulation.Nodes()
            ids = p_on_tri.Nodes()
            ids = [ids.Value(i) for i in range(ids.Lower(), ids.Upper() + 1)]

        if n > 1:
            idList = vtk.vtkIdList()
            for i in ids:
                pnt = pnts.Value(i)
                pnt.Transform(trsf)
                id = polydata.GetPoints().InsertNextPoint(pnt.X(), pnt.Y(), pnt.Z())
                idList.InsertNextId(id)
            polydata.InsertNextCell(vtk.VTK_POLY_LINE, idList)

    # 4. 转换Face并合并
    exp = TopExp_Explorer(sh, TopAbs_FACE)
    if exp.More():
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.stl'
            stlapi.Write(sh, str(p), False)

            r = vtk.vtkSTLReader()
            r.SetFileName(str(p))
            r.Update()

            a = vtk.vtkAppendPolyData()
            a.AddInputData(r.GetOutput())
            a.AddInputData(polydata)
            a.Update()

            c = vtk.vtkCleanPolyData()
            c.SetInputData(a.GetOutput())
            c.Update()

            polydata.DeepCopy(a.GetOutput())

    return polydata


def ccSh_Pd(pd: vtk.vtkPolyData):
    builder, compound = BRep_Builder(), TopoDS_Compound()
    builder.MakeCompound(compound)

    triangles = vtk.vtkTriangleFilter()
    triangles.SetInputData(pd)
    triangles.Update()
    triangles = triangles.GetOutput()

    for i in range(triangles.GetNumberOfCells()):
        cell: vtk.vtkCell = triangles.GetCell(i)
        if cell.GetCellType() == vtk.VTK_TRIANGLE:
            ids: vtk.vtkIdList = cell.GetPointIds()
            pts = [triangles.GetPoint(ids.GetId(0)),
                   triangles.GetPoint(ids.GetId(1)),
                   triangles.GetPoint(ids.GetId(2))]
            face = ccFace(ccWire(ccSegments(pts, True)))
            builder.Add(compound, face)

    sewing = BRepBuilderAPI_Sewing()
    sewing.Load(compound)
    sewing.Perform()

    sh = sewing.SewedShape()

    if sh.ShapeType() is TopAbs_SHELL and sh.Closed():
        return ccSolid(sh)
    else:
        return sh


def ccPnt(arg):
    return gp_Pnt(arg[0], arg[1], arg[2])


def ccVec(arg):
    return gp_Vec(arg[0], arg[1], arg[2])


def ccDir(arg):
    return gp_Dir(arg[0], arg[1], arg[2])


def ccAx2(origin, z, x=None):
    args = [ccPnt(origin), ccDir(z)]
    if x is not None:
        args.append(ccDir(x))
    return gp_Ax2(*args)


def ccSegment(pt0, pt1):
    pt0 = pt0 if isinstance(pt0, gp_Pnt) else ccPnt(pt0)
    pt1 = pt1 if isinstance(pt1, gp_Pnt) else ccPnt(pt1)
    return GC_MakeSegment(pt0, pt1).Value()


def ccSegments(pts, closed=False):
    pnts_ = [pnt for pnt in pts]
    if closed:
        pnts_.append(pts[0])
    return [ccSegment(pnts_[i], pnts_[i + 1]) for i in range(len(pnts_) - 1)]


def ccBSpline(pts, tgs=None, tgflags=None, closed=False, scale=True, tol=precision.Confusion()):
    if len(pts) < 2:
        return None

    if pts is not None and not isinstance(pts, TColgp_HArray1OfPnt):
        _ = TColgp_HArray1OfPnt(1, len(pts))
        for i in range(len(pts)):
            if isinstance(pts[i], gp_Pnt):
                _.SetValue(i + 1, pts[i])
            else:
                _.SetValue(i + 1, ccPnt(pts[i]))
        pts = _

    if tgs is not None and not isinstance(tgs, TColgp_Array1OfVec):
        _ = TColgp_Array1OfVec(1, len(tgs))
        for i in range(len(tgs)):
            if isinstance(pts[i], gp_Vec):
                _.SetValue(i + 1, tgs[i])
            else:
                _.SetValue(i + 1, ccVec(tgs[i]))
        tgs = _

    if tgs is not None and tgflags is None:
        tgflags = [True for _ in tgs]

    if tgflags is not None and not isinstance(tgflags, TColStd_HArray1OfBoolean):
        _ = TColStd_HArray1OfBoolean(1, len(tgflags))
        for i in range(len(tgflags)):
            _.SetValue(i + 1, tgflags[i])
        tgflags = _

    lower, upper = pts.Lower(), pts.Upper() + (1 if closed else 0)
    knots = TColStd_HArray1OfReal(lower, upper)
    knots.SetValue(lower, 0)

    for i in range(lower, upper):
        if closed and i == upper - 1:
            var = pts.Value(i).Distance(pts.Value(lower)) ** 0.5
        else:
            var = pts.Value(i).Distance(pts.Value(i + 1)) ** 0.5
        knots.SetValue(i + 1, knots.Value(i) + var)

    curve = GeomAPI_Interpolate(pts, knots, closed, tol)
    if tgs is not None:
        curve.Load(tgs, tgflags, scale)
    curve.Perform()
    return curve.Curve() if curve.IsDone() else None


def ccBSpline_Kochanek(pts, closed=False, tol=precision.Confusion()):
    if len(pts) < 2:
        return None

    points = vtk.vtkPoints()
    for pt in pts:
        points.InsertNextPoint(pt)

    ks = [vtk.vtkKochanekSpline(), vtk.vtkKochanekSpline(), vtk.vtkKochanekSpline()]

    ps = vtk.vtkParametricSpline()
    ps.SetXSpline(ks[0])
    ps.SetYSpline(ks[1])
    ps.SetZSpline(ks[2])
    ps.SetPoints(points)
    ps.SetClosed(1 if closed else 0)
    ps.SetParameterizeByLength(1)
    ps.SetLeftConstraint(1)
    ps.SetRightConstraint(1)

    n = 100 * len(pts)

    pf = vtk.vtkParametricFunctionSource()
    pf.SetParametricFunction(ps)
    pf.SetUResolution(n)
    pf.SetVResolution(n)
    pf.SetWResolution(n)
    pf.Update()

    pts_, points = [], pf.GetOutput().GetPoints()

    for i in range(points.GetNumberOfPoints() - (1 if closed else 0)):
        pts_.append(points.GetPoint(i))
    pts_[0] = pts[0].copy()
    if not closed:
        pts_[-1] = pts[-1].copy()

    curve = ccBSpline(pts_, closed=closed, tol=tol)

    obb = vmi.pdOBB_Pts(points)
    if 0 < obb['size'][2] < tol:
        curve = ccCurveOnPlane(curve, obb['center'], obb['axis'][2])
    return curve


def ccCurveOnPlane(curve, origin, normal, vt=None):
    if vt is None:
        vt = normal
    return geomprojlib.ProjectOnPlane(curve, ccPlane(origin, normal), ccDir(vt), True)


def ccCircle_Ax2(ax2, radius):
    mk = GC_MakeCircle(ax2, radius)
    return mk.Value() if mk.IsDone() else None


def ccCircle_CS(center, radius, cs: vmi.CS4x4):
    ax2 = ccAx2(center, cs.axis(2), cs.axis(0))
    return ccCircle_Ax2(ax2, radius)


def ccPolygonRegular_CS(center, radius, n, cs: vmi.CS4x4):
    pts, center, cs = [], np.array(center[:3]), cs.copy()
    cs.translate(radius * cs.axis(0))

    for i in range(n):
        cs.rotate(360 / n, cs.axis(2))
        pt = center + radius * cs.axis(0)
        pts.append(pt)

    return ccSegments(pts, True)


def ccPlane(origin, normal):
    mk = GC_MakePlane(ccPnt(origin), ccDir(normal))
    return mk.Value() if mk.IsDone() else None


def ccVertex(pt):
    pt = pt if isinstance(pt, gp_Pnt) else ccPnt(pt)
    mk = BRepBuilderAPI_MakeVertex(pt)
    mk.Build()
    return mk.Vertex() if mk.IsDone() else None


def ccEdge(curve):
    mk = BRepBuilderAPI_MakeEdge(curve)
    mk.Build()
    return mk.Edge() if mk.IsDone() else None


def ccEdges(curves):
    return [ccEdge(curve) for curve in curves]


def ccWire(edges):
    edges_ = edges
    if not hasattr(edges_, '__iter__'):
        edges_ = [edges_]
    edges_ = [edge if isinstance(edge, TopoDS_Edge) else ccEdge(edge) for edge in edges_]
    mk = BRepBuilderAPI_MakeWire()
    for edge in edges_:
        mk.Add(edge)
    mk.Build()
    return mk.Wire() if mk.IsDone() else None


def ccWires(edges):
    return [ccWire(edge) for edge in edges]


def ccFace(wire):
    mk = BRepBuilderAPI_MakeFace(wire)
    mk.Build()
    return mk.Face() if mk.IsDone() else None


def ccFaceOnPlane(wire, origin, normal):
    mk = BRepBuilderAPI_MakeFace(gp_Pln(ccPnt(origin), ccDir(normal)), wire)
    mk.Build()
    return mk.Face() if mk.IsDone() else None


def ccLoft(sections: List[TopoDS_Wire], ruled=False, redirect=False):
    mk = BRepOffsetAPI_ThruSections(False, ruled)
    mk.CheckCompatibility(redirect)
    for section in sections:
        if isinstance(section, TopoDS_Edge):
            section = ccWire(section)
        if isinstance(section, TopoDS_Wire):
            mk.AddWire(section)
        if isinstance(section, TopoDS_Vertex):
            mk.AddVertex(section)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccSew(shs: List[TopoDS_Shape]):
    mk = BRepBuilderAPI_Sewing()
    for sh in shs:
        mk.Add(sh)
    mk.Perform()
    return mk.SewedShape()


def ccSolid(shells):
    if isinstance(shells, TopoDS_Shell):
        shells = [shells]
    mk = BRepBuilderAPI_MakeSolid()
    for shell in shells:
        if shell.ShapeType() is TopAbs_SHELL:
            mk.Add(topods.Shell(shell))
    if mk.IsDone():
        s = mk.Solid()

        if ccInside_Pt(s):
            s.Reverse()
        return s


def ccBoolean_Union(sh: List[TopoDS_Shape]):
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Fuse()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Union', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccBoolean_Difference(sh: List[TopoDS_Shape]):
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Cut()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Difference', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccBoolean_Intersection(sh: List[TopoDS_Shape]):
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Common()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Intersection', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccSection(sh: List[TopoDS_Shape]):
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Section()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccSection', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccCylinder(radius, height, origin, z, x=None):
    mk = BRepPrimAPI_MakeCylinder(ccAx2(origin, z, x), radius, height)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccPrism(sh, vt, l):
    vt = np.array(vt)
    vec = l * vt / np.linalg.norm(vt)
    vec = gp_Vec(vec[0], vec[1], vec[2])
    mk = BRepPrimAPI_MakePrism(sh, vec)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccOffset(sh, offset):
    mk = BRepOffsetAPI_MakeOffsetShape(sh, offset, precision.Confusion())
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccRayCast(sh, pt, vt):
    """把pt沿vt方向投射到sh上"""
    pt, vt = np.array(pt), np.array(vt)

    def dmax(s0, s1):
        b = [Bnd_Box(), Bnd_Box()]
        brepbndlib.Add(s0, b[0])
        brepbndlib.Add(s1, b[1])

        min = [[b[0].Get()[i] for i in range(3)],
               [b[1].Get()[i] for i in range(3)]]
        max = [[b[0].Get()[i] for i in range(3, 6)],
               [b[1].Get()[i] for i in range(3, 6)]]
        xyz = [[gp_XYZ(min[0][0], min[0][1], min[0][2]),
                gp_XYZ(max[0][0], max[0][1], max[0][2])],
               [gp_XYZ(min[1][0], min[1][1], min[1][2]),
                gp_XYZ(max[1][0], max[1][1], max[1][2])]]
        delta = [xyz[0][1].Subtracted(xyz[0][0]),
                 xyz[1][1].Subtracted(xyz[1][0])]

        def dmin(s0, s1):
            b = [Bnd_Box(), Bnd_Box()]
            brepbndlib.Add(s0, b[0])
            brepbndlib.Add(s1, b[1])
            return b[0].Distance(b[1])

        return delta[0].Modulus() + delta[1].Modulus() + dmin(s0, s1)

    d = dmax(ccVertex(pt), sh)
    vtmax = d * vt / np.linalg.norm(vt)

    sweep = BRepSweep_Prism(ccVertex(pt), vtmax)
    section = ccSection([sh, sweep.Shape()])

    origin = ccPnt(pt)

    pts = {}
    for s in ccExplore_Recursive(section, TopAbs_VERTEX):
        pnt = BRep_Tool.Pnt(s)
        pts[pnt.Distance(origin)] = [pnt.X(), pnt.Y(), pnt.Z()]
    return [pts[i] for i in sorted(pts)]


def ccUniformPts_Edge(edge, n, u=None):
    if n < 2:
        return
    curve = BRepAdaptor_Curve(edge)
    mk = GCPnts_QuasiUniformAbscissa(curve, n)
    if mk.IsDone():
        pts = [curve.Value(mk.Parameter(i)) for i in range(1, n + 1)]
        return [[pt.X(), pt.Y(), pt.Z()] for pt in pts]


def ccEdge_FirstLastVertex(edge):
    pts = [TopoDS_Vertex(), TopoDS_Vertex()]
    topexp.Vertices(edge, pts[0], pts[1])
    return pts


def ccEdge_FirstLastPoint(edge):
    pts = ccEdge_FirstLastVertex(edge)
    pts = [BRep_Tool.Pnt(pt) for pt in pts]
    pts = [[pt.X(), pt.Y(), pt.Z()] for pt in pts]
    return pts


def ccLength(sh):
    p = GProp_GProps()
    brepgprop.LinearProperties(sh, p)
    return p.Mass()


def ccArea(sh):
    p = GProp_GProps()
    brepgprop.SurfaceProperties(sh, p)
    return p.Mass()


def ccInside_Pt(sh, pt=None):
    c = BRepClass3d_SolidClassifier(sh)
    if pt is None:
        c.PerformInfinitePoint(precision.Confusion())
    else:
        c.Perform(ccPnt(pt), precision.Confusion())

    if c.State() == TopAbs_IN or c.State() == TopAbs_ON:
        return True
    return False



import importlib
import inspect
import pkgutil
import PySide2
import OCC
import vtk
import CGAL


def walk_packages(p):
    err_starts, err_ends = ['_',
                            'vtk.gtk',
                            'vtk.qt.QVTKRenderWindowInteractor',
                            'vtk.qt4',
                            'vtk.tk'], ['Python']

    modules = set()
    modules_error = set()
    modules.add(p)

    for pkg in [_ for _ in pkgutil.walk_packages(p.__path__, p.__name__ + '.')]:
        err_flag = False

        if not err_flag:
            for err_start in err_starts:
                if pkg.name.startswith(err_start):
                    err_flag = True
                    break

        if not err_flag:
            for err_end in err_ends:
                if pkg.name.endswith(err_end):
                    err_flag = True
                    break

        if not err_flag:
            try:
                modules.add(importlib.import_module(pkg.name))
            except ImportError as err:
                modules_error.add(pkg.name)

    # print(len(modules))
    m_count = c_count = f_count = 0
    c_set = set()
    f_set = set()
    txt_file = open(p.__name__ + '.txt', 'w')

    for m in modules:
        m_count += 1
        m_str = m.__name__
        txt_file.write(m_str + '\t->\t' + m.__class__.__name__ + '\n')

        for c_name, c in inspect.getmembers(m, inspect.isclass):
            if hasattr(m, c_name):
                c_count += 1
                c_str = m_str + '.' + c_name
                c_set.add(c_name)
                # txt_file.write(c_str + '\t->\t' + c.__class__.__name__ + '\n')

                for f_name, f in inspect.getmembers(c):
                    f_count += 1
                    f_str = c_str + '.' + f_name
                    f_set.add(f_name)
                    # txt_file.write(f_str + '\t->\t' + f.__class__.__name__ + '\n')

    txt_file.close()
    print(p.__name__)
    print('module:', m_count)
    print('class:', c_count, len(c_set))
    print('member:', f_count, len(f_set))
    # print(modules_error)


if __name__ == '__main__':
    walk_packages(CGAL)
    # walk_packages(PySide2)
    # walk_packages(OCC)
    # walk_packages(vtk)
    # walk_packages(itk)



import os
import pathlib
import shutil
import sys
import tempfile
import threading

import SimpleITK as sitk
import pydicom
import vtk

from PySide2.QtCore import *
import vmi


def sortSeries(dcmdir, recursive=True):
    dcmdir = str(pathlib.Path(dcmdir))
    series = []

    def func():
        roots = {dcmdir}
        if recursive:
            roots = {root for root, *_ in os.walk(dcmdir)}
        for root in roots:
            with tempfile.TemporaryDirectory() as p:
                if vmi.diffEncode(root):
                    for f in os.listdir(root):
                        f = os.path.join(root, f)
                        if os.path.isfile(f):
                            shutil.copy2(f, p)
                    root = p

                for i in sitk.ImageSeriesReader.GetGDCMSeriesIDs(root):
                    filenames = sitk.ImageSeriesReader.GetGDCMSeriesFileNames(root, i)
                    series.append(SeriesData(filenames))

    t = threading.Thread(target=func)
    t.start()
    vmi.appwait(t)
    return series


class SeriesData(QObject, vmi.Menu):
    def __init__(self, filenames=None, name=None):
        QObject.__init__(self)

        self.name = name if name else self.tr('系列 （Series)')
        vmi.Menu.__init__(self, name=self.name)

        self._Filenames = filenames
        self._Directory = tempfile.TemporaryDirectory()

        if self._Filenames and len(self._Filenames) > 0:
            self._Filenames = [shutil.copy2(f, self._Directory.name) for f in self._Filenames]

    def __setstate__(self, s):
        self.__init__(name=s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Filenames = []
        self._Directory = tempfile.TemporaryDirectory()
        for kw in s['_Bytes']['_Files']:
            p = pathlib.Path(self._Directory.name) / kw
            p.write_bytes(s['_Bytes']['_Files'][kw])
            self._Filenames.append(str(p))

        del s['_Bytes']

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['_Filenames', '_Directory', 'menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Bytes'] = {'_Files': {pathlib.Path(f).name: pathlib.Path(f).read_bytes() for f in self._Filenames}}
        return s

    def __str__(self):
        with pydicom.dcmread(self._Filenames[0]) as ds:
            r = ['', '', '', '', '']
            for e in [_ for _ in ds if _.VR != 'SQ']:
                e.showVR = False
                if e.name in ['Modality', 'Slice Thickness', 'Rows', 'Columns', 'Pixel Spacing', 'Pixel Data']:
                    r[0] += repr(e) + '\n'
                elif e.name.startswith('Patient'):
                    r[1] += repr(e) + '\n'
                elif e.name.startswith('Study'):
                    r[2] += repr(e) + '\n'
                elif e.name.startswith('Series'):
                    r[3] += repr(e) + '\n'
                elif e.name.startswith('Image'):
                    r[4] += repr(e) + '\n'
            return 'slices: {}\n'.format(len(self._Filenames)) + r[0] + r[1] + r[2] + r[3] + r[4]

    def __repr__(self):
        return self.__str__()

    def filenames(self):
        return self._Filenames

    def readITK(self):
        r = sitk.ImageSeriesReader()
        r.SetFileNames(self._Filenames)
        return r.Execute()

    def read(self):
        image = None

        def func():
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.nii'

                w = sitk.ImageFileWriter()
                w.SetFileName(str(p))
                w.SetImageIO('NiftiImageIO')
                w.Execute(self.readITK())

                r = vtk.vtkNIFTIImageReader()
                r.SetFileName(str(p))
                r.Update()

                nonlocal image
                image = r.GetOutput()

        t = threading.Thread(target=func)
        t.start()
        vmi.appwait(t)
        return image


if __name__ == '__main__':
    import vmi.example.read_dicom_slice_view
    vmi.example.read_dicom_slice_view.run()



import tempfile
import pathlib
import math
from typing import Union, Optional, List

import SimpleITK as sitk
import vtk
import vmi
import numpy as np
import numba

UC = sitk.sitkUInt8
SS = sitk.sitkInt16
SI = sitk.sitkInt32
F = sitk.sitkFloat32
D = sitk.sitkFloat64


def imSITKScalar_VTK(t: int) -> int:
    """
    转换图像体素类型

    :param t: UC - unsigned char, SS - signed short, SI - signed int, F - float, D - double
    :return: vtk.VTK_UNSIGNED_CHAR, vtk.VTK_SHORT, vtk.VTK_INT, vtk.VTK_FLOAT, vtk.VTK_DOUBLE
    """
    return {vtk.VTK_UNSIGNED_CHAR: UC,
            vtk.VTK_SHORT: SS,
            vtk.VTK_INT: SI,
            vtk.VTK_FLOAT: F,
            vtk.VTK_DOUBLE: D}[t]


def imVTKScalar_SITK(t: int) -> int:
    """
    转换图像体素类型

    :param t: vtk.VTK_UNSIGNED_CHAR, vtk.VTK_SHORT, vtk.VTK_INT, vtk.VTK_FLOAT, vtk.VTK_DOUBLE
    :return: UC - unsigned char, SS - signed short, SI - signed int, F - float, D - double
    """
    return {UC: vtk.VTK_UNSIGNED_CHAR,
            SS: vtk.VTK_SHORT,
            SI: vtk.VTK_INT,
            F: vtk.VTK_FLOAT,
            D: vtk.VTK_DOUBLE}[t]


def imSITK_VTK(image: Union[vtk.vtkImageData, sitk.Image], output_scalar_type=SS) -> sitk.Image:
    """
    转换图像数据类型

    :param image: 输入vtkImageData
    :param output_scalar_type: 输出图像的体素类型
    :return: 输出SimpleITK.Image
    """
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.nii'

        if isinstance(image, vtk.vtkImageData):
            w = vtk.vtkNIFTIImageWriter()
            w.SetFileName(str(p))
            w.SetInputData(image)
            w.Update()
        else:
            w = sitk.ImageFileWriter()
            w.SetFileName(str(p))
            w.SetImageIO('NiftiImageIO')
            w.Execute(image)

        r = sitk.ImageFileReader()
        r.SetFileName(str(p))
        r.SetImageIO('NiftiImageIO')
        r.SetOutputPixelType(output_scalar_type)
        return r.Execute()


def imVTK_SITK(image: Union[sitk.Image, vtk.vtkImageData], data_scalar_type=vtk.VTK_INT) -> vtk.vtkImageData:
    """
    转换图像数据类型

    :param image: 输入SimpleITK.Image
    :param data_scalar_type: 输出图像的体素类型
    :return: 输出vtkImageData
    """
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.nii'

        if isinstance(image, vtk.vtkImageData):
            w = vtk.vtkNIFTIImageWriter()
            w.SetFileName(str(p))
            w.SetInputData(image)
            w.Update()
        else:
            w = sitk.ImageFileWriter()
            w.SetFileName(str(p))
            w.SetImageIO('NiftiImageIO')
            w.Execute(image)

        r = vtk.vtkNIFTIImageReader()
        r.SetFileName(str(p))
        r.SetDataScalarType(data_scalar_type)
        r.Update()
        return r.GetOutput()


def imArray_VTK(image: vtk.vtkImageData) -> np.ndarray:
    """
    转换图像数据类型

    :param image: 输入vtkImageData
    :return: 输出ndarray
    """
    return sitk.GetArrayFromImage(imSITK_VTK(image))


def imVTK_Array(ary: np.ndarray, origin: np.ndarray, spacing: np.ndarray) -> vtk.vtkImageData:
    """
    转换图像数据类型

    :param origin: 图像原点
    :param spacing: 图像体素间距
    :param extent: 图像数据坐标范围
    :param ary: 输入图像体素值ndarray
    :return: 输出图像vtkImageData
    """
    if ary.ndim == 3:
        ext = [0, ary.shape[2] - 1, 0, ary.shape[1] - 1, 0, ary.shape[0] - 1]
    elif ary.ndim == 2:
        ext = [0, ary.shape[1] - 1, 0, ary.shape[0] - 1, 0, 0]
    elif ary.ndim == 1:
        ext = [0, ary.shape[0] - 1, 0, 0, 0, 0]
    else:
        ext = [0, 0, 0, 0, 0, 0]

    image = imVTK_SITK(sitk.GetImageFromArray(ary))
    image.SetOrigin(origin[0], origin[1], origin[2])
    image.SetSpacing(spacing[0], spacing[1], spacing[2])
    image.SetExtent(ext[0], ext[1], ext[2], ext[3], ext[4], ext[5])
    return image


def imOrigin(image: vtk.vtkImageData) -> np.ndarray:
    """返回图像原点[x, y, z]"""
    return np.array(image.GetOrigin())


def imBounds(image: vtk.vtkImageData) -> np.ndarray:
    """返回图像边界[xmin, xmax, ymin, ymax, zmin, zmax]"""
    return np.array(image.GetBounds())


def imCenter(image: vtk.vtkImageData) -> np.ndarray:
    """返回图像中心点[x, y, z]"""
    return np.array(image.GetCenter())


def imDimensions(image: vtk.vtkImageData) -> np.ndarray:
    """返回图像数据维数(整数)"""
    return np.array(image.GetDimensions())


def imExtent(image: vtk.vtkImageData, bounds: Union[np.ndarray, List[float]] = None) -> np.ndarray:
    """
    返回图像数据范围(整数)，如果设置了边界范围bounds则输出边界内的数据范围

    :param image: 输入图像
    :param bounds: 边界范围[xmin, xmax, ymin, ymax, zmin, zmax]
    :return: 数据范围[imin, imax, jmin, jmax, kmin, kmax]
    """
    if bounds is not None:
        ori, spc, ext = imOrigin(image), imSpacing(image), imExtent(image)
        first, last = [bounds[0], bounds[2], bounds[4]], [bounds[1], bounds[3], bounds[5]]
        for i in range(3):
            f = int((first[i] - ori[i]) / spc[i])
            l = int((last[i] - ori[i]) / spc[i] + 1)
            if f > ext[2 * i]:
                ext[2 * i] = f
            if l < ext[2 * i + 1]:
                ext[2 * i + 1] = l
        return ext
    return np.array(image.GetExtent())


def imSpacing(image: vtk.vtkImageData) -> np.ndarray:
    """
    返回图像沿三个坐标轴方向的体素间距

    :param image: 输入图像
    :return: [spacing_x, spacing_y, spacing_z]
    """
    return np.array(image.GetSpacing())


def imSpacing_Vt(image: vtk.vtkImageData, vt: np.ndarray = None) -> float:
    """
    返回图像的体素间距，如果设置了方向vt则返回沿该方向的体素间距

    :param image: 输入图像
    :param vt: 任意方向[x, y, z]
    :return: 各方向体素间距[dx, dy, dz]，或沿vt方向体素间距dvt
    """
    spc, vt = imSpacing(image), np.array([vt[0], vt[1], vt[2]])
    spc = vmi.vtOnVector(spc, vt)
    return np.linalg.norm(spc)


def imSize(image: vtk.vtkImageData) -> np.ndarray:
    """
    返回图像包围盒沿三个坐标轴方向的尺度

    :param image: 输入图像
    :return: [size_x, size_y, size_z]
    """
    bnd = imBounds(image)
    return np.array([bnd[1] - bnd[0], bnd[3] - bnd[2], bnd[5] - bnd[4]])


def imSize_Vt(image: vtk.vtkImageData, vt: np.ndarray) -> float:
    """
    返回图像包围盒沿方向vt的尺度

    :param image: 输入图像
    :param vt: 任意方向[x, y, z]
    :return: 各方向图像包围和边长[sx, sy, sz]，或沿vt方向体素间距svt
    """
    size, vt = imSize(image), np.array([vt[0], vt[1], vt[2]])
    size = vmi.vtOnVector(size, vt)
    return np.linalg.norm(size)


def imStencil(pd: vtk.vtkPolyData, output_origin: np.ndarray, output_spacing: np.ndarray,
              output_extent: np.ndarray) -> vtk.vtkImageStencilData:
    """
    将面网格数据pd转化为图像模板数据vtkImageStencilData

    :param pd: 输入面网格数据
    :param output_extent: 输出数据范围[imin, imax, jmin, jmax, kmin, kmax]
    :param output_spacing: 输出体素间距[dx, dy, dz]
    :param output_origin: 输出图像原点[x, y, z]
    :return: 图像模板数据vtkImageStencilData
    """
    stencil = vtk.vtkPolyDataToImageStencil()
    stencil.SetInputData(pd)
    stencil.SetOutputOrigin([output_origin[0], output_origin[1], output_origin[2]])
    stencil.SetOutputSpacing([output_spacing[0], output_spacing[1], output_spacing[2]])
    stencil.SetOutputWholeExtent([output_extent[0], output_extent[1], output_extent[2]])
    stencil.Update()
    return stencil.GetOutput()


def imValue(image: vtk.vtkImageData, ijk: List[int] = None, value: float = None, component: int = 0) -> float:
    """
    获得体素值

    :param image: 输入图像
    :param ijk: 体素值数据坐标
    :param value: 改写体素值
    :param component: 体素值是向量时的元素位置，默认为0
    :return: 体素值
    """
    if value is not None:
        image.SetScalarComponentFromDouble(ijk[0], ijk[1], ijk[2], component, value)
        image.Modified()
    return image.GetScalarComponentAsDouble(ijk[0], ijk[1], ijk[2], component)


def imResample_Isotropic(image: vtk.vtkImageData) -> vtk.vtkImageData:
    """
    各项同性重采样，将图像各个方向不均匀的体素间距均匀化，输出立方体体素体积不变

    :param image: 输入图像
    :return: 输出重采样图像
    """
    spc = imSpacing(image)
    isospc = (spc[0] * spc[1] * spc[2]) ** (1 / 3)

    itkimage: sitk.Image = imSITK_VTK(image, F)

    smooth = sitk.RecursiveGaussianImageFilter()
    smooth.SetDirection(0)
    smooth.SetSigma(isospc)
    itkimage = smooth.Execute(itkimage)

    smooth = sitk.RecursiveGaussianImageFilter()
    smooth.SetDirection(1)
    smooth.SetSigma(isospc)
    itkimage = smooth.Execute(itkimage)

    size = itkimage.GetSize()
    dx = size[0] * spc[0] / isospc
    dy = size[1] * spc[1] / isospc
    dz = size[2] * spc[2] / isospc

    resample = sitk.ResampleImageFilter()
    resample.SetTransform(sitk.Transform())
    resample.SetInterpolator(sitk.sitkLinear)
    resample.SetDefaultPixelValue(-3071)
    resample.SetOutputSpacing([isospc, isospc, isospc])
    resample.SetOutputOrigin(itkimage.GetOrigin())
    resample.SetOutputDirection(itkimage.GetDirection())
    resample.SetSize([int(dx), int(dy), int(dz)])
    itkimage = resample.Execute(itkimage)

    itkimage = imSITK_VTK(itkimage, imSITKScalar_VTK(image.GetScalarType()))
    return imVTK_SITK(itkimage)


def imIsosurface(image: vtk.vtkImageData, isovalue: Union[int, float]) -> vtk.vtkPolyData:
    """
    提取图像等值面三维重建为面网格数据
    
    :param image: 输入图像
    :param isovalue: 等体素值
    :return: 三维面网格数据
    """
    fe3 = vtk.vtkFlyingEdges3D()
    fe3.SetInputData(image)
    fe3.SetValue(0, isovalue)
    fe3.SetComputeNormals(1)
    fe3.SetComputeGradients(0)
    fe3.SetComputeScalars(0)
    fe3.Update()
    return fe3.GetOutput()


def imReslice(image: vtk.vtkImageData,
              cs: Optional[vmi.CS4x4] = None,
              size: Optional[List[float]] = None,
              scalar_background: Union[int, float] = 0) -> vtk.vtkImageData:
    """
    图像重采样

    :param image: 输入图像
    :param cs: 输出图像的坐标系
    :param size: 输出图像的尺度
    :param scalar_background: 输出体素的背景值
    :return: 输出重采样图像
    """
    if cs is None:
        cs = vmi.CS4x4()

    if size is None:
        size = [vmi.imSize_Vt(image, cs.axis(0)),
                vmi.imSize_Vt(image, cs.axis(1)),
                vmi.imSize_Vt(image, cs.axis(2))]

    spc = [vmi.imSpacing_Vt(image, cs.axis(0)),
           vmi.imSpacing_Vt(image, cs.axis(1)),
           vmi.imSpacing_Vt(image, cs.axis(2))]

    ext = [0, math.ceil(size[0] / spc[0]) - 1,
           0, math.ceil(size[1] / spc[1]) - 1,
           0, math.ceil(size[2] / spc[2]) - 1]

    reslice = vtk.vtkImageReslice()
    reslice.SetInputData(image)
    reslice.SetInterpolationModeToCubic()
    reslice.SetResliceAxes(cs.vtkmat4x4())
    reslice.SetOutputOrigin([0, 0, 0])
    reslice.SetOutputSpacing(spc)
    reslice.SetOutputExtent(ext)
    reslice.SetBackgroundLevel(scalar_background)
    reslice.Update()
    return reslice.GetOutput()


def imReslice_Blend(image: vtk.vtkImageData,
                    cs: Optional[vmi.CS4x4] = None,
                    size: Optional[List[float]] = None,
                    scalar_background: Union[int, float] = 0,
                    scalar_mode: str = 'mean',
                    scalar_min: Union[int, float] = None,
                    scalar_max: Union[int, float] = None) -> vtk.vtkImageData:
    """
    图像重采样，并沿输出坐标系的z轴方向混叠

    :param image: 输入图像
    :param cs: 输出图像的坐标系
    :param size: 输出图像的尺度
    :param scalar_background: 输出体素的背景值
    :param scalar_mode: mean - 平均体素值（默认） max - 最大体素值 min - 最小体素值
    :param scalar_min: 混叠体素最小值
    :param scalar_max: 混叠体素最大值
    :return: 输出重采样图像
    """

    @numba.jit(nopython=True)
    def blend(ary: np.ndarray, scalar_min=None, scalar_max=None):
        ary = ary.copy()

        # 沿z方向累加大于minimum的体素，取平均值
        for j in range(ary.shape[1]):
            for i in range(ary.shape[2]):
                voxels = []
                for k in range(ary.shape[0]):
                    if scalar_min is not None and ary[k][j][i] < scalar_min:
                        continue
                    elif scalar_max is not None and ary[k][j][i] > scalar_max:
                        continue
                    else:
                        voxels.append(ary[k][j][i])

                if len(voxels):
                    if scalar_mode == 'mean':
                        ary[0][j][i] = np.mean(np.array(voxels))
                    elif scalar_mode == 'max':
                        ary[0][j][i] = np.max(np.array(voxels))
                    elif scalar_mode == 'min':
                        ary[0][j][i] = np.min(np.array(voxels))
                    else:
                        ary[0][j][i] = np.mean(np.array(voxels))
                else:
                    ary[0][j][i] = scalar_background

        return ary[0]

    image = imReslice(image, cs, size, scalar_background)
    ary = vmi.imArray_VTK(image)
    ary = blend(ary, scalar_min, scalar_max)

    image = vmi.imVTK_Array(ary, vmi.imOrigin(image), vmi.imSpacing(image))

    return image
